import errno
import logging
import multiprocessing
import os
import pickle
import sys
from pathlib import Path
from subprocess import CalledProcessError

import numpy as np
import pytest
from numpy.testing import assert_equal

import nengo
import nengo.neurons
import nengo.utils.least_squares_solvers
from nengo.cache import (
    CacheIndex,
    CacheIOError,
    DecoderCache,
    Fingerprint,
    NoDecoderCache,
    WriteableCacheIndex,
    get_fragment_size,
    safe_stat,
)
from nengo.exceptions import CacheIOWarning, FingerprintError
from nengo.solvers import LstsqL2


class Mock:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    def __getitem__(self, key):
        return Mock()

    def __iter__(self):
        return iter([])

    def __mul__(self, other):
        return 1.0

    @classmethod
    def __getattr__(cls, name):
        if name in ("__file__", "__path__"):
            return "/dev/null"
        if name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        return Mock()


def isstrictsubclass(x, cls):
    return isinstance(x, type) and issubclass(x, cls) and x is not cls


def list_objects(module):
    return [getattr(module, key) for key in dir(module) if not key.startswith("_")]


neuron_types = [
    obj
    for obj in list_objects(nengo.neurons)
    if isstrictsubclass(obj, nengo.neurons.NeuronType)
    and obj is not nengo.neurons.RatesToSpikesNeuronType
]
solver_types = [
    obj
    for obj in list_objects(nengo.solvers)
    if isstrictsubclass(obj, nengo.solvers.Solver)
    and obj is not nengo.solvers.NoSolver  # NoSolver is not cacheable
]
lstsq_solver_types = [
    obj
    for obj in list_objects(nengo.utils.least_squares_solvers)
    if isstrictsubclass(obj, nengo.utils.least_squares_solvers.LeastSquaresSolver)
]


class SolverMock:
    n_calls = {}

    def __init__(self):
        self.n_calls[self] = 0

    def __call__(self, conn, gain, bias, x, targets, rng=np.random, E=None):
        self.n_calls[self] += 1
        if E is None:
            return np.random.rand(x.shape[1], targets.shape[1]), {"info": "v"}
        return np.random.rand(x.shape[1], E.shape[1]), {"info": "v"}


def get_solver_test_args(
    n_eval_points=100,
    n_neurons=10,
    dimensions=2,
    solver=nengo.solvers.LstsqL2nz(),
    weights=False,
):
    conn = nengo.Connection(
        nengo.Ensemble(n_neurons, dimensions, add_to_container=False),
        nengo.Node(size_in=dimensions, add_to_container=False),
        solver=solver,
        add_to_container=False,
    )
    defaults = {
        "conn": conn,
        "gain": np.ones(n_neurons),
        "bias": np.ones(n_neurons),
        "x": np.ones((n_eval_points, dimensions)),
        "targets": np.ones((n_eval_points, n_neurons)),
        "rng": np.random.RandomState(42),
    }
    if weights:
        defaults["E"] = np.ones((dimensions, n_neurons // 2))
    return defaults


def test_decoder_cache(tmp_path):
    cache_dir = str(tmp_path)

    # Basic test, that results are cached.
    with DecoderCache(cache_dir=cache_dir) as cache:
        solver_mock = SolverMock()
        decoders1, solver_info1 = cache.wrap_solver(solver_mock)(
            **get_solver_test_args()
        )
        assert SolverMock.n_calls[solver_mock] == 1
        decoders2, solver_info2 = cache.wrap_solver(solver_mock)(
            **get_solver_test_args()
        )
        assert SolverMock.n_calls[solver_mock] == 1  # result read from cache?
        assert_equal(decoders1, decoders2)
        assert solver_info1 == solver_info2

        solver_args = get_solver_test_args()
        solver_args["gain"] *= 2
        decoders3, _ = cache.wrap_solver(solver_mock)(**solver_args)
        assert SolverMock.n_calls[solver_mock] == 2
        assert np.any(decoders1 != decoders3)

        # Test that the cache does not load results of another solver.
        another_solver = SolverMock()
        cache.wrap_solver(another_solver)(
            **get_solver_test_args(solver=nengo.solvers.LstsqNoise())
        )
        assert SolverMock.n_calls[another_solver] == 1

    with DecoderCache(cache_dir=cache_dir, readonly=True) as cache:
        n_calls = int(SolverMock.n_calls[solver_mock])
        decoders3, solver_info3 = cache.wrap_solver(solver_mock)(
            **get_solver_test_args()
        )
        assert SolverMock.n_calls[solver_mock] == n_calls
        assert_equal(decoders3, decoders2)
        assert solver_info3 == solver_info2


def test_corrupted_decoder_cache(tmp_path):
    cache_dir = str(tmp_path)

    with DecoderCache(cache_dir=cache_dir) as cache:
        solver_mock = SolverMock()
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1

        # corrupt the cache
        for path in cache.get_files():
            with open(path, "w", encoding="utf-8") as f:
                f.write("corrupted")

        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 2


def test_corrupted_decoder_cache_index(tmp_path):
    cache_dir = str(tmp_path)

    with DecoderCache(cache_dir=cache_dir):
        pass  # Initialize cache with required files
    assert len(os.listdir(cache_dir)) == 2  # index, index.lock

    # Write corrupted index
    with open(os.path.join(cache_dir, CacheIndex._INDEX), "w", encoding="utf-8") as f:
        f.write("(d")  # empty dict, but missing '.' at the end

    # Try to load index
    with DecoderCache(cache_dir=cache_dir):
        pass
    assert len(os.listdir(cache_dir)) == 2  # index, index.lock


def test_too_new_decoder_cache_index(tmp_path):
    cache_dir = str(tmp_path)

    # Write index with super large version numbers
    with open(os.path.join(cache_dir, CacheIndex._INDEX), "wb") as f:
        pickle.dump((1000, 1000), f)

    with pytest.warns(UserWarning, match="could not acquire lock and was deactivated"):
        with DecoderCache(cache_dir=cache_dir) as cache:
            solver_mock = SolverMock()
            cache.wrap_solver(solver_mock)(**get_solver_test_args())
            assert SolverMock.n_calls[solver_mock] == 1


def test_decoder_cache_invalidation(tmp_path):
    cache_dir = str(tmp_path)
    solver_mock = SolverMock()

    # Basic test, that results are cached.
    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1
        cache.invalidate()
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 2


def test_readonly_cache(caplog, tmp_path):
    caplog.set_level(logging.INFO)
    cache_dir = str(tmp_path)

    with open(os.path.join(cache_dir, CacheIndex._INDEX), "wb") as f:
        pickle.dump((CacheIndex.VERSION, pickle.HIGHEST_PROTOCOL), f)
        pickle.dump({}, f)

    with DecoderCache(readonly=True, cache_dir=cache_dir) as cache:
        cache.shrink()
        with pytest.raises(CacheIOError, match="Cannot invalidate a readonly cache."):
            cache.invalidate()
    assert len(caplog.records) == 1
    assert caplog.records[0].message == "Tried to shrink a readonly cache."


def test_decoder_cache_size_includes_overhead(tmp_path):
    cache_dir = str(tmp_path)
    solver_mock = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())

        fragment_size = get_fragment_size(cache_dir)
        actual_size = sum(os.stat(p).st_size for p in cache.get_files())
        assert actual_size % fragment_size != 0, (
            "Test succeeded by chance. Adjust get_solver_test_args() to "
            "produce date not aligned with the files system fragment size."
        )

        assert cache.get_size_in_bytes() % fragment_size == 0


def test_decoder_cache_shrinking(tmp_path):
    cache_dir = str(tmp_path)
    solver_mock = SolverMock()
    another_solver = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())

        # Ensure differing time stamps (depending on the file system the
        # timestamp resolution might be as bad as 1 day).
        for path in cache.get_files():
            timestamp = os.stat(path).st_atime
            timestamp -= 60 * 60 * 24 * 2  # 2 days
            os.utime(path, (timestamp, timestamp))

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(another_solver)(
            **get_solver_test_args(solver=nengo.solvers.LstsqNoise())
        )

        cache_size = cache.get_size_in_bytes()
        assert cache_size > 0

        cache.shrink(cache_size - 1)

        # check that older cached result was removed
        assert SolverMock.n_calls[solver_mock] == 1
        cache.wrap_solver(another_solver)(
            **get_solver_test_args(solver=nengo.solvers.LstsqNoise())
        )
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 2
        assert SolverMock.n_calls[another_solver] == 1


def test_decoder_cache_shrink_threadsafe(monkeypatch, tmp_path):
    """Tests that shrink handles files deleted by other processes."""
    cache_dir = str(tmp_path)
    solver_mock = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        limit = cache.get_size()

        # Ensure differing time stamps (depending on the file system the
        # timestamp resolution might be as bad as 1 day).
        for filename in os.listdir(cache.cache_dir):
            path = os.path.join(cache.cache_dir, filename)
            timestamp = os.stat(path).st_atime
            timestamp -= 60 * 60 * 24 * 2  # 2 days
            os.utime(path, (timestamp, timestamp))

        cache.wrap_solver(solver_mock)(
            **get_solver_test_args(solver=nengo.solvers.LstsqNoise())
        )

        cache_size = cache.get_size_in_bytes()
        assert cache_size > 0

        def raise_file_not_found(orig_fn):
            def fn(filename, *args, **kwargs):
                if filename.endswith(".lock"):
                    return orig_fn(filename, *args, **kwargs)
                raise OSError(errno.ENOENT, "File not found.")

            return fn

        monkeypatch.setattr(cache, "get_size_in_bytes", lambda: cache_size)
        monkeypatch.setattr("os.stat", raise_file_not_found(os.stat))
        monkeypatch.setattr("os.remove", raise_file_not_found(os.remove))
        monkeypatch.setattr("os.unlink", raise_file_not_found(os.unlink))

        cache.shrink(limit)


def test_decoder_cache_with_E_argument_to_solver(tmp_path):
    cache_dir = str(tmp_path)
    solver_mock = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        decoders1, solver_info1 = cache.wrap_solver(solver_mock)(
            **get_solver_test_args(weights=True)
        )
        assert SolverMock.n_calls[solver_mock] == 1
        decoders2, solver_info2 = cache.wrap_solver(solver_mock)(
            **get_solver_test_args(weights=True)
        )
        assert SolverMock.n_calls[solver_mock] == 1  # read from cache?
        assert_equal(decoders1, decoders2)
        assert solver_info1 == solver_info2


class DummyA:
    def __init__(self, attr=0):
        self.attr = attr


nengo.cache.Fingerprint.whitelist(DummyA)


class DummyB:
    def __init__(self, attr=0):
        self.attr = attr


nengo.cache.Fingerprint.whitelist(DummyB, fn=lambda obj: True)


def dummy_fn(arg):
    pass


@pytest.mark.parametrize(
    "reference, equal, different",
    (
        (True, True, False),  # bool
        (False, False, True),  # bool
        (1.0, 1.0, 2.0),  # float
        (1, 1, 2),  # int
        (1.0 + 2.0j, 1 + 2j, 2.0 + 1j),  # complex
        (b"a", b"a", b"b"),  # bytes
        ((0, 1), (0, 1), (0, 2)),  # tuple
        ([0, 1], [0, 1], [0, 2]),  # list
        ("a", "a", "b"),  # str
        (np.eye(2), np.eye(2), np.array([[0, 1], [1, 0]])),  # array
        (DummyA(), DummyA(), DummyB()),  # object instance
        (DummyA(1), DummyA(1), DummyA(2)),  # object instance
        (LstsqL2(reg=0.1), LstsqL2(reg=0.1), LstsqL2(reg=0.2)),  # solver
    ),
)
def test_fingerprinting(reference, equal, different):
    assert str(Fingerprint(reference)) == str(Fingerprint(equal))
    assert str(Fingerprint(reference)) != str(Fingerprint(different))


@pytest.mark.parametrize(
    "obj",
    (
        np.array([object()]),  # array
        np.array([(1.0,)], dtype=[("field1", "f8")]),  # array
        object(),  # object instance
        dummy_fn,  # function
    ),
)
def test_unsupported_fingerprinting(obj):
    with pytest.raises(FingerprintError, match="cannot be fingerprinted"):
        Fingerprint(obj)


@pytest.mark.parametrize("cls", neuron_types + solver_types + lstsq_solver_types)
def test_supported_fingerprinting(cls, monkeypatch):
    # patch so we can instantiate various solvers without the proper libraries
    monkeypatch.setitem(sys.modules, "scipy", Mock())
    monkeypatch.setitem(sys.modules, "scipy.optimize", Mock())
    monkeypatch.setitem(sys.modules, "scipy.sparse", Mock())
    monkeypatch.setitem(sys.modules, "scipy.sparse.linalg", Mock())
    monkeypatch.setitem(sys.modules, "sklearn", Mock())
    monkeypatch.setitem(sys.modules, "sklearn.linear_model", Mock())
    monkeypatch.setitem(sys.modules, "sklearn.utils", Mock())
    monkeypatch.setitem(sys.modules, "sklearn.utils.extmath", Mock())

    args = []
    if issubclass(cls, nengo.neurons.RatesToSpikesNeuronType):
        # spiking types require a `base_type` argument, so provide one
        args.append(nengo.neurons.RectifiedLinear())

    obj = cls(*args)
    assert Fingerprint.supports(obj)

    # check fingerprint is created without error and is a valid sha1 hash
    fp = str(Fingerprint(obj))
    assert len(fp) == 40 and set(fp).issubset("0123456789abcdef")


def test_fails_for_lambda_expression():
    with pytest.raises(FingerprintError, match="cannot be fingerprinted"):
        Fingerprint(lambda x: x)


def test_cache_works(tmp_path, Simulator, seed):
    cache_dir = str(tmp_path)

    model = nengo.Network(seed=seed)
    with model:
        nengo.Connection(nengo.Ensemble(10, 1), nengo.Ensemble(10, 1))

    assert len(os.listdir(cache_dir)) == 0
    with Simulator(
        model,
        model=nengo.builder.Model(
            dt=0.001, decoder_cache=DecoderCache(cache_dir=cache_dir)
        ),
    ):
        assert len(os.listdir(cache_dir)) == 3  # index, index.lock, and *.nco


def test_cache_not_used_without_seed(tmp_path, Simulator):
    cache_dir = str(tmp_path)

    model = nengo.Network()
    with model:
        nengo.Connection(nengo.Ensemble(10, 1), nengo.Ensemble(10, 1))

    assert len(os.listdir(cache_dir)) == 0
    with Simulator(
        model,
        model=nengo.builder.Model(
            dt=0.001, decoder_cache=DecoderCache(cache_dir=cache_dir)
        ),
    ):
        assert len(os.listdir(cache_dir)) == 2  # index, index.lock


def build_many_ensembles(cache_dir, Simulator):
    with nengo.Network(seed=1) as model:
        for _ in range(100):
            nengo.Connection(nengo.Ensemble(10, 1), nengo.Ensemble(10, 1))

    with Simulator(
        model,
        model=nengo.builder.Model(
            dt=0.001, decoder_cache=DecoderCache(cache_dir=cache_dir)
        ),
    ):
        pass


@pytest.mark.slow
def test_cache_concurrency(tmp_path, Simulator):
    cache_dir = str(tmp_path)

    n_processes = 100
    processes = [
        multiprocessing.Process(
            target=build_many_ensembles, args=(cache_dir, Simulator)
        )
        for _ in range(n_processes)
    ]
    for p in processes:
        p.start()
    for p in processes:
        p.join(60)
    for p in processes:
        assert p.exitcode == 0


def test_warns_out_of_context(tmp_path):
    cache_dir = str(tmp_path)
    cache = DecoderCache(cache_dir=cache_dir)

    solver_mock = SolverMock()
    solver = cache.wrap_solver(solver_mock)
    with pytest.warns(UserWarning):
        solver(**get_solver_test_args())
    assert SolverMock.n_calls[solver_mock] == 1


def test_cacheindex_cannot_write(tmp_path):
    index = WriteableCacheIndex(cache_dir=str(tmp_path))
    with index:
        index[0] = ("file0", 0, 0)
    mtime = os.stat(index.index_path).st_mtime

    index = CacheIndex(cache_dir=str(tmp_path))
    with index:
        with pytest.raises(TypeError):
            index[0] = ("file", 0, 0)
        with pytest.raises(TypeError):
            del index[0]
        assert index[0] == ("file0", 0, 0)
    assert os.stat(index.index_path).st_mtime == mtime


def test_writeablecacheindex_writes(tmp_path):
    index = WriteableCacheIndex(cache_dir=str(tmp_path))
    with index:
        index[0] = ("file0", 0, 0)
        index[1] = ("file1", 0, 0)
        del index[1]

    # Verify with readonly cacheindex
    index = CacheIndex(cache_dir=str(tmp_path))
    with index:
        assert index[0] == ("file0", 0, 0)
        assert 1 not in index


def test_writeablecacheindex_setitem(tmp_path):
    index = WriteableCacheIndex(cache_dir=str(tmp_path))

    with pytest.raises(ValueError):
        index[0] = "file0"
    with pytest.raises(ValueError):
        index[0] = ("file0", 0)
    with pytest.raises(ValueError):
        index[0] = ("file0", 0, 0, 0)


def test_writeablecacheindex_removes(tmp_path):
    index = WriteableCacheIndex(cache_dir=str(tmp_path))
    with index:
        index[0] = ("file0", 0, 0)
        index[1] = ("file1", 0, 0)
        index.remove_file_entry("file0")
        index.remove_file_entry(os.path.join(str(tmp_path), "file1"))

    # Verify with readonly cacheindex
    index = CacheIndex(cache_dir=str(tmp_path))
    with index:
        assert 0 not in index, "Fails on relative paths"
        assert 1 not in index, "Fails on absolute paths"


def test_writeablecacheindex_warning(monkeypatch, tmp_path):
    with pytest.warns(CacheIOWarning):
        with WriteableCacheIndex(cache_dir=str(tmp_path)) as idx:

            class RaiseError(type(idx.cache_dir)):
                def replace(self, *args, **kwargs):
                    raise CalledProcessError(-1, "move")

            monkeypatch.setattr(idx, "cache_dir", RaiseError(idx.cache_dir))


def test_writeablecacheindex_reinit(tmp_path):
    with WriteableCacheIndex(cache_dir=str(tmp_path)) as idx:
        subdir = idx.cache_dir / "subdir"
        subdir.mkdir()
        file1 = idx.cache_dir / "file1"
        file2 = subdir / "file2"
        with file1.open("w", encoding="utf-8") as fh:
            fh.write("contents1")
        with file2.open("w", encoding="utf-8") as fh:
            fh.write("contents2")

        assert subdir.exists() and file1.exists() and file2.exists()

        idx._reinit()
        assert not (subdir.exists() or file1.exists() or file2.exists())


def test_shrink_does_not_fail_if_lock_cannot_be_acquired(tmp_path):
    cache = DecoderCache(cache_dir=str(tmp_path))
    cache._index._lock.timeout = 1.0
    with cache:
        cache.wrap_solver(SolverMock())(**get_solver_test_args())
    with cache._index._lock:
        cache.shrink(limit=0)


def test_safe_stat(caplog):
    assert safe_stat(Path("/tmp/does/not/exist")) is None
    assert len(caplog.records) == 1
    assert "OSError during safe_stat:" in caplog.records[0].message


def test_no_decoder_cache():
    cache = NoDecoderCache()
    assert cache.get_size_in_bytes() == 0
    assert cache.get_size() == "0 B"



import collections

from nengo.utils import graphs


def graph(edges=None):
    """Construct graph from edges."""
    g = collections.defaultdict(set)
    if edges is not None:
        g.update(edges)
    return g


def test_reversedict():
    edges = graph({"a": {"b", "c"}, "b": set(), "c": set()})
    r_edges = graphs.reverse_edges(edges)
    assert r_edges == {"a": set(), "b": {"a"}, "c": {"a"}}


def test_toposort():
    edges = graph({"a": {"b", "c"}, "b": {"c"}, "c": set()})
    assert graphs.toposort(edges) == ["a", "b", "c"]


def test_transitive_closure():
    edges = graph({"a": {}, "b": {"c", "d"}, "c": set(), "d": {"e"}, "e": set()})
    assert graphs.transitive_closure(edges) == {
        "a": set(),
        "b": {"c", "d", "e"},
        "c": set(),
        "d": {"e"},
        "e": set(),
    }


def test_add_edges():
    edges = graph({"a": {"b", "c"}})
    graphs.add_edges(edges, [("a", "d"), ("b", "c")])
    assert edges == {"a": {"b", "c", "d"}, "b": {"c"}}



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

import nengo
from nengo.builder.neurons import SimNeurons
from nengo.builder.operator import Copy, ElementwiseInc
from nengo.builder.optimizer import (
    CopyMerger,
    ElementwiseIncMerger,
    OpMerger,
    OpsToMerge,
    SigMerger,
    SimNeuronsMerger,
)
from nengo.builder.signal import Signal
from nengo.tests.test_learning_rules import learning_net
from nengo.transforms import SparseMatrix


def test_sigmerger_check():
    # 0-d signals
    assert SigMerger.check([Signal(0), Signal(0)])
    assert not SigMerger.check([Signal(0), Signal(1)])

    # compatible along first axis
    assert SigMerger.check([Signal(shape=(1, 2)), Signal(shape=(2, 2))])

    # compatible along second axis
    assert SigMerger.check([Signal(shape=(2, 1)), Signal(shape=(2, 2))], axis=1)
    assert not SigMerger.check([Signal(shape=(2, 1)), Signal(shape=(2, 2))], axis=0)

    # shape mismatch
    assert not SigMerger.check([Signal(shape=(2,)), Signal(shape=(2, 2))])

    # mixed dtype
    assert not SigMerger.check(
        [Signal(np.zeros(2, dtype=int)), Signal(np.zeros(2, dtype=float))]
    )

    s1 = Signal(shape=(5,))
    s2 = Signal(shape=(5,))

    # mixed signal and view
    assert not SigMerger.check([s1, s1[:3]])

    # mixed bases
    assert not SigMerger.check([s1[:2], s2[2:]])

    # compatible views
    assert SigMerger.check([s1[:2], s1[2:]])

    # sparse signals not mergeable
    assert not SigMerger.check(
        [
            Signal(SparseMatrix([[0, 0]], 1.0, (1, 1))),
            Signal(SparseMatrix([[0, 0]], 1.0, (1, 1))),
        ]
    )

    # same signal cannot appear twice
    sig = Signal(0)
    assert not SigMerger.check([sig, sig])


def test_sigmerger_check_signals():
    # 0-d signals
    SigMerger.check_signals([Signal(0), Signal(0)])
    with pytest.raises(ValueError):
        SigMerger.check_signals([Signal(0), Signal(1)])

    # compatible along first axis
    SigMerger.check_signals([Signal(shape=(1, 2)), Signal(shape=(2, 2))])

    # compatible along second axis
    SigMerger.check_signals([Signal(shape=(2, 1)), Signal(shape=(2, 2))], axis=1)
    with pytest.raises(ValueError):
        SigMerger.check_signals([Signal(shape=(2, 1)), Signal(shape=(2, 2))], axis=0)

    # shape mismatch
    with pytest.raises(ValueError):
        SigMerger.check_signals([Signal(shape=(2,)), Signal(shape=(2, 2))])

    # mixed dtype
    with pytest.raises(ValueError):
        SigMerger.check_signals(
            [Signal(np.zeros(2, dtype=int)), Signal(np.zeros(2, dtype=float))]
        )

    # compatible views
    s = Signal(shape=(5,))
    with pytest.raises(ValueError):
        SigMerger.check_signals([s[:2], s[2:]])


def test_sigmerger_check_views():
    s1 = Signal(shape=(5, 5))
    s2 = Signal(shape=(5, 5))

    # compatible along first axis
    SigMerger.check_views([s1[:1], s1[1:]])

    # compatible along second axis
    SigMerger.check_views([s1[:1, :1], s1[:1, 1:]], axis=1)

    # non-views
    with pytest.raises(ValueError, match="Cannot merge non-views"):
        SigMerger.check_views([s1, s2])

    # different bases
    with pytest.raises(ValueError, match="must share the same base"):
        SigMerger.check_views([s1[:2], s2[2:]])

    # different ndims
    with pytest.raises(ValueError, match="must have the same number of dimensions"):
        SigMerger.check_views([s1[:1], s1[1:, 0]])

    # different strides
    with pytest.raises(ValueError, match="must have equal strides"):
        SigMerger.check_views([s1[::2], s1[::3]])

    # shape mismatch
    with pytest.raises(ValueError, match="must have same shape except on"):
        SigMerger.check_views([s1[:1, :1], s1[:1, 1:]], axis=0)

    # non-sequential
    with pytest.raises(ValueError, match="Views are not sequential"):
        SigMerger.check_views([s1[1:], s1[1:-1]])


def test_sigmerger_merge(allclose):
    s1 = Signal(np.array([[0, 1], [2, 3]]))
    s2 = Signal(np.array([[4, 5]]))

    sig, replacements = SigMerger.merge([s1, s2])
    assert allclose(sig.initial_value, np.array([[0, 1], [2, 3], [4, 5]]))
    assert allclose(replacements[s1].initial_value, s1.initial_value)
    assert allclose(replacements[s2].initial_value, s2.initial_value)

    with pytest.raises(ValueError, match="Cannot merge mixed views and non-views"):
        SigMerger.merge([s1[0], s2])


def test_sigmerger_merge_views(allclose):
    s = Signal(np.array([[0, 1], [2, 3], [4, 5]]))
    v1, v2 = s[:2], s[2:]
    merged, _ = SigMerger.merge_views([v1, v2])

    assert allclose(merged.initial_value, s.initial_value)
    assert v1.base is s
    assert v2.base is s


def test_opstomerge_check_signals():
    sig = Signal(np.arange(10))
    sig_orig = sig.reshape(10)
    sig_reshaped = sig.reshape(2, 5)
    assert not OpsToMerge.check_signals(
        Copy(sig_orig, sig_orig), Copy(sig_reshaped, sig_reshaped)
    )


def test_opmerger_warning():
    with pytest.warns(UserWarning, match="Merger for operator type"):
        OpMerger.register(Copy)(CopyMerger)


def test_copymerger_merge_slice():
    sig1 = Signal(np.ones(2))
    slice1 = [1]
    sig2 = Signal(np.ones(3))
    with pytest.raises(ValueError, match="Mixed Ellipsis with list of indices."):
        CopyMerger.merge_slice([sig1, sig2], [slice1, None])
    merged = CopyMerger.merge_slice([sig1, sig2], [slice1, [0, 2]])
    assert merged == [1, 2, 4]


def test_elementwiseincmerger_scalars():
    y1 = Signal(shape=(1,))
    y2 = Signal(shape=(1,))
    a = Signal(shape=(1,))
    x1 = Signal(shape=(1,))
    x2 = Signal(shape=(1,))

    inc1 = ElementwiseInc(a, x1, y1)
    inc2 = ElementwiseInc(a, x2, y2)

    assert ElementwiseIncMerger.is_mergeable(inc1, inc2)
    merged_inc, _ = ElementwiseIncMerger.merge([inc1, inc2])
    assert merged_inc.Y.shape == (2,)
    assert merged_inc.Y.name.startswith("merged")
    assert merged_inc.A.shape == (1,)
    assert merged_inc.A is a
    assert merged_inc.X.shape == (2,)
    assert merged_inc.X.name.startswith("merged")


def test_simneuronsmerger_warning(rng):
    nt = nengo.PoissonSpiking(nengo.Tanh())
    op1 = SimNeurons(
        nt, J=Signal(shape=(1,)), output=Signal(shape=(1,)), state={"rng": rng}
    )

    op2 = SimNeurons(
        nt, J=Signal(shape=(1,)), output=Signal(shape=(1,)), state={"rng": rng}
    )
    assert SimNeuronsMerger.is_mergeable(op1, op2)
    with pytest.warns(UserWarning, match="Extra state has been modified"):
        SimNeuronsMerger.merge([op1, op2])


def test_optimizer_does_not_change_result(seed):
    dtype_resolution = np.finfo(nengo.rc.float_dtype).resolution
    dtype_decimal = int(np.floor(-np.log10(dtype_resolution) * 0.5))

    model = learning_net()
    model.seed = seed

    with model:
        # Add the default probe for every non-Probe object
        probes = [
            nengo.Probe(obj)
            for obj in model.all_objects
            if not isinstance(obj, (nengo.Probe, nengo.Network))
        ]
        # Also probe learning rules and neurons
        probes.extend(nengo.Probe(ens.neurons) for ens in model.all_ensembles)
        probes.extend(
            nengo.Probe(conn.learning_rule)
            for conn in model.all_connections
            if conn.learning_rule is not None
        )

    with nengo.Simulator(model, optimize=False) as sim:
        sim.run(0.1)
    with nengo.Simulator(model, optimize=True) as sim_opt:
        sim_opt.run(0.1)

    for probe in probes:
        assert_almost_equal(sim.data[probe], sim_opt.data[probe], decimal=dtype_decimal)



"""
Vendorized from https://github.com/renmengye/np-conv2d

MIT License

Copyright (c) 2017 Mengye Ren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import division

import numpy as np

from nengo.utils.numpy import array_offset


def calc_pad(pad, in_siz, out_siz, stride, ksize):
    """Calculate padding width.
    Args:
        pad: padding method, "SAME", "VALID", or manually speicified.
        ksize: kernel size [I, J].
    Returns:
        pad_: Actual padding width.
    """
    if pad == 'SAME':
        return max((out_siz - 1) * stride + ksize - in_siz, 0)
    elif pad == 'VALID':
        return 0
    else:
        return pad


def calc_gradx_pad(pad, in_siz, out_siz, stride, ksize):
    if pad == 'SAME':
        out_siz_min = (in_siz - 1) * stride + 1
        p = out_siz + ksize - 1 - out_siz_min
        p = max(p, 0)
        p = min(p, (ksize - 1) * 2)
        return p
    elif pad == 'VALID':
        return (ksize - 1) * 2
    else:
        return pad


def calc_size(h, kh, pad, sh):
    """Calculate output image size on one dimension.
    Args:
        h: input image size.
        kh: kernel size.
        pad: padding strategy.
        sh: stride.
    Returns:
        s: output size.
    """

    if pad == 'VALID':
        return np.ceil((h - kh + 1) / sh)
    elif pad == 'SAME':
        return np.ceil(h / sh)
    else:
        return int(np.ceil((h - kh + pad + 1) / sh))


def extract_sliding_windows(x, ksize, pad, stride, floor_first=True):
    """Converts a tensor to sliding windows.
    Args:
        x: [N, H, W, C]
        k: [KH, KW]
        pad: [PH, PW]
        stride: [SH, SW]
    Returns:
        y: [N, (H-KH+PH+1)/SH, (W-KW+PW+1)/SW, KH * KW, C]
    """
    n, h, w, c = x.shape
    kh, kw = ksize
    sh, sw = stride

    h2 = int(calc_size(h, kh, pad, sh))
    w2 = int(calc_size(w, kw, pad, sw))
    ph = int(calc_pad(pad, h, h2, sh, kh))
    pw = int(calc_pad(pad, w, w2, sw, kw))

    ph0 = int(np.floor(ph / 2))
    ph1 = int(np.ceil(ph / 2))
    pw0 = int(np.floor(pw / 2))
    pw1 = int(np.ceil(pw / 2))

    if floor_first:
        pph = (ph0, ph1)
        ppw = (pw0, pw1)
    else:
        pph = (ph1, ph0)
        ppw = (pw1, pw0)
    x = np.pad(
        x, ((0, 0), pph, ppw, (0, 0)),
        mode='constant',
        constant_values=(0.0,))

    x_sn, x_sh, x_sw, x_sc = x.strides
    y_strides = (x_sn, sh*x_sh, sw*x_sw, x_sh, x_sw, x_sc)
    y = np.ndarray((n, h2, w2, kh, kw, c), dtype=x.dtype, buffer=x.data,
                   offset=array_offset(x), strides=y_strides)
    return y


def extract_sliding_windows_gradx(x,
                                  ksize,
                                  pad,
                                  stride,
                                  orig_size,
                                  floor_first=False):
    """Extracts windows on a dilated image.

    Args:
        x: [N, H', W', C] (usually dy)
        k: [KH, KW]
        pad: [PH, PW]
        stride: [SH, SW]
        orig_size: [H, W]

    Returns:
        y: [N, H, W, KH, KW, C]
    """
    n, h, w, c = x.shape
    kh, kw = ksize
    sh, sw = stride
    ph, pw = pad
    sh, sw = stride
    h2, w2 = orig_size

    xs = np.zeros([n, h, sh, w, sw, c])
    xs[:, :, 0, :, 0, :] = x
    xss = xs.shape
    x = xs.reshape([xss[0], xss[1] * xss[2], xss[3] * xss[4], xss[5]])
    x = x[:, :h2, :w2, :]

    ph2 = int(np.ceil(ph / 2))
    ph3 = int(np.floor(ph / 2))
    pw2 = int(np.ceil(pw / 2))
    pw3 = int(np.floor(pw / 2))
    if floor_first:
        pph = (ph3, ph2)
        ppw = (pw3, pw2)
    else:
        pph = (ph2, ph3)
        ppw = (pw2, pw3)
    x = np.pad(
        x, ((0, 0), pph, ppw, (0, 0)),
        mode='constant',
        constant_values=(0.0, ))

    # The following code extracts window without copying the data:
    # y = np.zeros([n, h2, w2, kh, kw, c])
    # for ii in range(h2):
    #     for jj in range(w2):
    #         y[:, ii, jj, :, :, :] = x[:, ii:ii + kh, jj:jj + kw, :]
    x_sn, x_sh, x_sw, x_sc = x.strides
    y_strides = (x_sn, x_sh, x_sw, x_sh, x_sw, x_sc)
    y = np.ndarray((n, h2, w2, kh, kw, c),
                   dtype=x.dtype,
                   buffer=x.data,
                   offset=array_offset(x),
                   strides=y_strides)
    return y


def conv2d(x, w, pad='SAME', stride=(1, 1)):
    """2D convolution (technically speaking, correlation).
    Args:
        x: [N, H, W, C]
        w: [I, J, C, K]
        pad: [PH, PW]
        stride: [SH, SW]
    Returns:
        y: [N, H', W', K]
    """
    ksize = w.shape[:2]
    x = extract_sliding_windows(x, ksize, pad, stride)
    ws = w.shape
    w = w.reshape([ws[0] * ws[1] * ws[2], ws[3]])
    xs = x.shape
    x = x.reshape([xs[0] * xs[1] * xs[2], -1])
    y = x.dot(w)
    y = y.reshape([xs[0], xs[1], xs[2], -1])
    return y


def conv2d_gradx(w, dy, xsize, pad='SAME', stride=(1, 1)):
    """2D convolution gradient wrt. input.

    Args:
        dy: [N, H', W', K]
        w: [I, J, K, C]
        xsize: Original image size, [H, W]

    Returns:
        dx: [N, H, W, C]

    Modifications from original code:
      - Do not transpose input/output channels in `w`
      - Fix bug in computing `pad2w` (use `dys[1]` instead of `dys[0]`)
      - Add new `calc_gradx_pad` function to ensure we compute padding the same as TF

    These changes have been added as a PR on the upstream repo
    https://github.com/renmengye/np-conv2d/pull/2, once those are merged in we can
    switch this back to a direct copy.
    """
    assert w.shape[-2] == dy.shape[-1]

    dys = dy.shape[1:3]
    ksize = w.shape[:2]
    pad2 = (
        calc_gradx_pad(pad, dys[0], xsize[0], stride[0], ksize[0]),
        calc_gradx_pad(pad, dys[1], xsize[1], stride[1], ksize[1]),
    )

    dx = extract_sliding_windows_gradx(dy, ksize, pad2, stride, xsize)
    dxs = dx.shape
    dx = dx.reshape([dxs[0] * dxs[1] * dxs[2], -1])
    w = w[::-1, ::-1, :, :]
    ws = w.shape
    w = w.reshape([ws[0] * ws[1] * ws[2], ws[3]])
    dx = dx.dot(w)
    return dx.reshape([dxs[0], dxs[1], dxs[2], -1])



from io import StringIO

import numpy as np

import nengo.utils.numpy as npext
from nengo.exceptions import SignalError
from nengo.rc import rc
from nengo.transforms import SparseMatrix


def is_sparse(obj):
    """Check if ``obj`` is a sparse matrix."""
    return isinstance(obj, SparseMatrix) or npext.is_spmatrix(obj)


class Signal:
    """Represents data or views onto data within a Nengo simulation.

    Signals are tightly coupled to NumPy arrays, which is how live data is
    represented in a Nengo simulation. Signals provide a view onto the
    important metadata of the live NumPy array, and maintain the original
    value of the array in order to reset the simulation to the initial state.

    Parameters
    ----------
    initial_value : array_like
        The initial value of the signal. Much of the metadata tracked by the
        Signal is based on this array as well (e.g., dtype).
    name : str, optional
        Name of the signal. Primarily used for debugging.
        If None, the memory location of the Signal will be used.
    base : Signal, optional
        The base signal, if this signal is a view on another signal.
        Linking the two signals with the ``base`` argument is necessary
        to ensure that their live data is also linked.
    readonly : bool, optional
        Whether this signal and its related live data should be marked as
        readonly. Writing to these arrays will raise an exception.
    offset : int, optional
        For a signal view this gives the offset of the view from the base
        ``initial_value`` in bytes. This might differ from the offset
        of the NumPy array view provided as ``initial_value`` if the base
        is a view already (in which case the signal base offset will be 0
        because it starts where the view starts. That NumPy view can have
        an offset of itself).
    """

    # Set assert_named_signals True to raise an Exception
    # if model.signal is used to create a signal with no name.
    # This can help to identify code that's creating un-named signals,
    # if you are trying to track down mystery signals that are showing
    # up in a model.
    assert_named_signals = False

    def __init__(
        self,
        initial_value=None,
        shape=None,
        name=None,
        base=None,
        readonly=False,
        offset=0,
    ):
        if self.assert_named_signals:
            assert name is not None
        self._name = name

        if initial_value is None:
            assert shape is not None
            initial_value = np.zeros(shape, dtype=rc.float_dtype)
        elif shape is not None:
            assert initial_value.shape == shape

        self._initial_value = initial_value
        if np.any(
            np.isnan(self._initial_value.data if self.sparse else self._initial_value)
        ):
            raise SignalError(f"{self!r} contains NaNs.")

        if self.sparse:
            assert initial_value.ndim == 2
            assert offset == 0
            assert base is None
            if npext.is_spmatrix(initial_value):
                self._initial_value.data.setflags(write=False)
        else:
            # To ensure we do not modify data passed into the signal,
            # we make a view of the data and mark it as not writeable.
            # Consumers (like SignalDict) are responsible for making copies
            # that can be modified, or using the readonly view appropriately.
            readonly_view = np.asarray(self._initial_value)
            if readonly_view.ndim > 0 and base is None:
                readonly_view = np.ascontiguousarray(readonly_view)
            # Ensure we have a view and aren't modifying the original's flags
            readonly_view = readonly_view.view()
            readonly_view.setflags(write=False)
            self._initial_value = readonly_view

        if base is not None:
            assert isinstance(base, Signal) and not base.is_view
            # make sure initial_value uses the same data as base.initial_value
            assert initial_value.base is base.initial_value.base
        self._base = base
        self._offset = offset

        self._readonly = bool(readonly)

    def __getstate__(self):
        state = dict(self.__dict__)

        if not self.sparse:
            # For normal arrays, the initial value could be a view on another
            # signal's data. To make sure we do not make a copy of the data,
            # we store the underlying metadata in a tuple, which pickle can
            # inspect to see that v.base is the same in different signals
            # and avoid serializing it multiple times.
            v = self._initial_value
            state["_initial_value"] = (
                v.shape,
                v.base,
                npext.array_offset(v),
                v.strides,
            )

        return state

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)

        if not self.sparse:
            shape, base, offset, strides = self._initial_value
            self._initial_value = np.ndarray(
                shape, buffer=base, dtype=base.dtype, offset=offset, strides=strides
            )
            self._initial_value.setflags(write=False)

    def __getitem__(self, item):
        """Index or slice into array."""
        if self.sparse:
            raise SignalError("Attempting to create a view of a sparse Signal")

        if item is Ellipsis or (isinstance(item, slice) and item == slice(None)):
            return self

        if not isinstance(item, tuple):
            item = (item,)

        if not all(npext.is_integer(i) or isinstance(i, slice) for i in item):
            raise SignalError("Can only index or slice into signals")

        if all(npext.is_integer(i) for i in item):
            # turn one index into slice to get a view from numpy
            item = item[:-1] + (slice(item[-1], item[-1] + 1),)

        view = self._initial_value[item]
        offset = npext.array_offset(view) - npext.array_offset(self._initial_value)
        return Signal(
            view,
            name=f"{self.name}[{item}]",
            base=self.base,
            offset=offset,
            readonly=self.readonly,
        )

    def __repr__(self):
        return f"Signal(name={self._name}, shape={self.shape})"

    @property
    def base(self):
        """(Signal) The base signal, if this signal is a view.

        Linking the two signals with the ``base`` argument is necessary
        to ensure that their live data is also linked.
        """
        return self if self._base is None else self._base

    @property
    def dtype(self):
        """(numpy.dtype) Data type of the signal (e.g., float64)."""
        return self.initial_value.dtype

    @property
    def elemoffset(self):
        """(int) Offset of data from base in elements."""
        return self.offset // self.itemsize

    @property
    def elemstrides(self):
        """(int) Strides of data in elements."""
        return None if self.sparse else tuple(s // self.itemsize for s in self.strides)

    @property
    def initial_value(self):
        """(numpy.ndarray) Initial value of the signal.

        Much of the metadata tracked by the Signal is based on this array
        as well (e.g., dtype).
        """
        return self._initial_value

    @initial_value.setter
    def initial_value(self, val):
        raise SignalError("Cannot change initial value after initialization")

    @property
    def is_view(self):
        """(bool) True if this Signal is a view on another Signal."""
        return self._base is not None

    @property
    def itemsize(self):
        """(int) Size of an array element in bytes."""
        return self.dtype.itemsize

    @property
    def name(self):
        """(str) Name of the signal. Primarily used for debugging."""
        return self._name if self._name is not None else (f"0x{id(self):x}")

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def nbytes(self):
        """(int) Number of bytes consumed by the signal."""
        return self.itemsize * self.size

    @property
    def ndim(self):
        """(int) Number of array dimensions."""
        return self.initial_value.ndim

    @property
    def offset(self):
        """(int) Offset of data from base in bytes.

        For a signal view this gives the offset of the view from the base
        ``initial_value`` in bytes. This might differ from the offset
        of the NumPy array view provided as ``initial_value`` if the base
        is a view already (in which case the signal base offset will be 0
        because it starts where the view starts. That NumPy view can have
        an offset of itself).
        """
        return self._offset

    @property
    def readonly(self):
        """(bool) Whether associated live data can be changed."""
        return self._readonly

    @readonly.setter
    def readonly(self, readonly):
        self._readonly = bool(readonly)

    @property
    def shape(self):
        """(tuple) Tuple of array dimensions."""
        return self.initial_value.shape

    @property
    def size(self):
        """(int) Total number of elements."""
        return self.initial_value.size

    @property
    def strides(self):
        """(tuple) Strides of data in bytes."""
        return None if self.sparse else self.initial_value.strides

    @property
    def sparse(self):
        """(bool) Whether the signal is sparse."""
        return is_sparse(self.initial_value)

    def may_share_memory(self, other):
        """Determine if two signals might overlap in memory.

        This comparison is not exact and errs on the side of false positives.
        See `numpy.may_share_memory` for more details.

        Parameters
        ----------
        other : Signal
            The other signal we are investigating.
        """
        return (self is other) or (
            (self.is_view or other.is_view)
            and np.may_share_memory(self.initial_value, other.initial_value)
        )

    def reshape(self, *shape):
        """Return a view on this signal with a different shape.

        Note that ``reshape`` cannot change the overall size of the signal.
        See `numpy.reshape` for more details.

        Any number of integers can be passed to this method,
        describing the desired shape of the returned signal.
        """

        if self.sparse:
            raise SignalError("Attempting to create a view of a sparse Signal")

        if len(shape) == 1:
            shape = shape[0]  # in case a tuple is passed in
        initial_value = self.initial_value.view()
        try:
            # this raises AttributeError if cannot reshape without copying
            initial_value.shape = shape
        except AttributeError as e:
            raise SignalError(
                f"Reshaping {self} to {shape} would require the array to be copied "
                "(because it is not contiguous), which is not supported"
            ) from e
        return Signal(
            initial_value,
            name=f"{self.name}.reshape({shape})",
            base=self.base,
            offset=self.offset,
        )


class SignalDict(dict):
    """Map from Signal -> ndarray.

    This dict subclass ensures that the ndarray values aren't overwritten,
    and instead data are written into them, which ensures that
    these arrays never get copied, which wastes time and space.

    Use ``init`` to set the ndarray initially.
    """

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            if isinstance(key, Signal) and key.base is not key:
                # return a view on the base signal
                base = dict.__getitem__(self, key.base)
                retval = np.ndarray(
                    buffer=base,
                    dtype=key.dtype,
                    shape=key.shape,
                    offset=key.offset,
                    strides=key.strides,
                )
                dict.__setitem__(self, key, retval)
                return retval
            else:
                raise

    def __setitem__(self, key, val):
        """Ensures that ndarrays stay in the same place in memory.

        Unlike normal dicts, this means that you cannot add a new key
        to a SignalDict using __setitem__. This is by design, to avoid
        silent typos when debugging Simulator. Every key must instead
        be explicitly initialized with SignalDict.init.
        """
        self[key][...] = val

    def __str__(self):
        """Pretty-print the signals and current values."""
        sio = StringIO()
        for k in self:
            sio.write(f"{repr(k)} {repr(self[k])}\n")
        return sio.getvalue()

    def init(self, signal):
        """Set up a permanent mapping from signal -> data."""
        if signal in self:
            raise SignalError("Cannot add signal twice")

        assert isinstance(signal, Signal)
        data = signal.initial_value
        if isinstance(data, SparseMatrix):
            data = data.allocate()

        if signal.is_view:
            if signal.base not in self:
                self.init(signal.base)

            # get a view onto the base data
            view = np.ndarray(
                shape=data.shape,
                strides=data.strides,
                offset=signal.offset,
                dtype=data.dtype,
                buffer=self[signal.base].data,
            )
            assert np.array_equal(view, data)
            view.setflags(write=not signal.readonly)
            dict.__setitem__(self, signal, view)
        else:
            data = data if signal.readonly else data.copy()
            dict.__setitem__(self, signal, data)

    def reset(self, signal):
        """Reset ndarray to the base value of the signal that maps to it."""
        if not signal.readonly:
            self[signal] = signal.initial_value



import pickle
from copy import copy

import numpy as np
import pytest

import nengo
from nengo import spa
from nengo.exceptions import NetworkContextError, NotAddedToNetworkWarning
from nengo.params import IntParam, NdarrayParam, iter_params
from nengo.utils.numpy import is_array_like
from nengo.utils.progress import TerminalProgressBar


def assert_is_copy(cp, original):
    assert cp is not original  # ensures separate parameters
    for param in iter_params(cp):
        param_inst = getattr(cp, param)
        if isinstance(param_inst, (nengo.solvers.Solver, nengo.base.NengoObject)):
            assert param_inst is getattr(original, param)
        elif is_array_like(param_inst):
            assert np.all(param_inst == getattr(original, param))
        else:
            assert param_inst == getattr(original, param)


def assert_is_deepcopy(cp, original):
    assert cp is not original  # ensures separate parameters
    for param in iter_params(cp):
        param_inst = getattr(cp, param)
        if isinstance(param_inst, (nengo.solvers.Solver, nengo.base.NengoObject)):
            assert_is_copy(param_inst, getattr(original, param))
        elif is_array_like(param_inst):
            assert np.all(param_inst == getattr(original, param))
        else:
            assert param_inst == getattr(original, param)


def make_ensemble():
    with nengo.Network():
        e = nengo.Ensemble(10, 1, radius=2.0)
    return e


def make_probe():
    with nengo.Network():
        e = nengo.Ensemble(10, 1)
        p = nengo.Probe(e, synapse=0.01)
    return p


def make_node():
    with nengo.Network():
        n = nengo.Node(np.min, size_in=2, size_out=2)
    return n


def make_connection():
    with nengo.Network():
        e1 = nengo.Ensemble(10, 1)
        e2 = nengo.Ensemble(10, 1)
        c = nengo.Connection(e1, e2, transform=2.0)
    return c


def make_function_connection():
    with nengo.Network():
        e1 = nengo.Ensemble(10, 1)
        e2 = nengo.Ensemble(10, 1)
        c = nengo.Connection(e1, e2, function=lambda x: x ** 2)
    return c


def make_learning_connection():
    with nengo.Network():
        e1 = nengo.Ensemble(10, 1)
        e2 = nengo.Ensemble(11, 1)
        c = nengo.Connection(e1.neurons, e2.neurons, transform=np.ones((11, 10)))
        c.learning_rule_type = nengo.PES()
        nengo.Connection(e2, c.learning_rule)
    return c


def make_network():
    with nengo.Network() as model:
        e1 = nengo.Ensemble(10, 1)
        e2 = nengo.Ensemble(10, 1)
        nengo.Connection(e1, e2, transform=2.0)
        nengo.Probe(e2)
    return model


def test_neurons_reference_copy():
    original = make_ensemble()
    cp = original.copy(add_to_container=False)
    assert original.neurons.ensemble is original
    assert cp.neurons.ensemble is cp


def test_learningrule_reference_copy():
    original = make_learning_connection()
    cp = original.copy(add_to_container=False)
    assert original.learning_rule.connection is original
    assert cp.learning_rule.connection is cp


def test_copy_in_network_default_add():
    original = make_network()

    with nengo.Network() as model:
        cp = original.copy()
    assert cp in model.all_objects

    assert_is_deepcopy(cp, original)


def test_copy_outside_network_default_add():
    original = make_network()
    cp = original.copy()
    assert_is_deepcopy(cp, original)


def test_network_copies_defaults():
    original = nengo.Network()
    original.config[nengo.Ensemble].radius = 1.5
    original.config[nengo.Connection].synapse = nengo.Lowpass(0.1)

    cp = original.copy()
    assert cp.config[nengo.Ensemble].radius == 1.5
    assert cp.config[nengo.Connection].synapse == nengo.Lowpass(0.1)


def test_network_copy_allows_independent_manipulation():
    with nengo.Network() as original:
        nengo.Ensemble(10, 1)
    original.config[nengo.Ensemble].radius = 2.0

    cp = original.copy()
    with cp:
        e2 = nengo.Ensemble(10, 1)
    cp.config[nengo.Ensemble].radius = 1.0

    assert e2 not in original.ensembles
    assert original.config[nengo.Ensemble].radius == 2.0


def test_copies_structure():
    with nengo.Network() as original:
        e1 = nengo.Ensemble(10, 1)
        e2 = nengo.Ensemble(10, 1)
        nengo.Connection(e1, e2)
        nengo.Probe(e2)

    cp = original.copy()

    assert cp.connections[0].pre is not e1
    assert cp.connections[0].post is not e2
    assert cp.connections[0].pre in cp.ensembles
    assert cp.connections[0].post in cp.ensembles

    assert cp.probes[0].target is not e2
    assert cp.probes[0].target in cp.ensembles


def test_network_copy_builds(Simulator):
    with Simulator(make_network().copy()):
        pass


def test_copy_obj_view():
    with nengo.Network():
        ens = nengo.Ensemble(10, 4)
        original = ens[:2]

    cp = original.copy()

    assert cp is not original
    assert cp.obj is ens
    assert cp.slice == original.slice


def test_copy_obj_view_in_connection():
    with nengo.Network() as original:
        node = nengo.Node([0.1, 0.2])
        ens = nengo.Ensemble(10, 2)
        nengo.Connection(node[0], ens[1])
        nengo.Connection(node[1], ens[0])

    cp = original.copy()
    assert cp.nodes[0] is not node
    assert cp.ensembles[0] is not ens
    assert cp.connections[0].pre.obj is cp.nodes[0]
    assert cp.connections[1].pre.obj is cp.nodes[0]
    assert cp.connections[0].post.obj is cp.ensembles[0]
    assert cp.connections[1].post.obj is cp.ensembles[0]


def test_pickle_obj_view():
    with nengo.Network():
        ens = nengo.Ensemble(10, 4)
        original = ens[:2]

    cp = pickle.loads(pickle.dumps(original))

    assert cp is not original
    assert cp.obj is not ens
    assert cp.obj.n_neurons == 10
    assert cp.obj.dimensions == 4
    assert cp.slice == original.slice


def test_pickle_obj_view_in_connection():
    with nengo.Network() as original:
        node = nengo.Node([0.1, 0.2])
        ens = nengo.Ensemble(10, 2)
        nengo.Connection(node[0], ens[1])
        nengo.Connection(node[1], ens[0])

    cp = pickle.loads(pickle.dumps(original))
    assert cp.nodes[0] is not node
    assert cp.ensembles[0] is not ens
    assert cp.connections[0].pre.obj is cp.nodes[0]
    assert cp.connections[1].pre.obj is cp.nodes[0]
    assert cp.connections[0].post.obj is cp.ensembles[0]
    assert cp.connections[1].post.obj is cp.ensembles[0]


@pytest.mark.parametrize(
    ("make_f", "assert_f"),
    [
        (make_ensemble, assert_is_copy),
        (make_probe, assert_is_copy),
        (make_node, assert_is_copy),
        (make_connection, assert_is_copy),
        (make_function_connection, assert_is_copy),
        (make_learning_connection, assert_is_copy),
        (make_network, assert_is_copy),
    ],
)
class TestCopy:
    """A basic set of tests that should pass for all objects."""

    def test_copy_in_network(self, make_f, assert_f):
        original = make_f()

        with nengo.Network() as model:
            cp = original.copy(add_to_container=True)
        assert cp in model.all_objects

        assert_f(cp, original)

    def test_copy_in_network_without_adding(self, make_f, assert_f):
        original = make_f()

        with nengo.Network() as model:
            cp = original.copy(add_to_container=False)
        assert cp not in model.all_objects

        assert_f(cp, original)

    def test_copy_outside_network(self, make_f, assert_f):
        original = make_f()
        with pytest.raises(NetworkContextError, match="either be created inside a"):
            original.copy(add_to_container=True)

    def test_copy_outside_network_without_adding(self, make_f, assert_f):
        original = make_f()
        cp = original.copy(add_to_container=False)
        assert_f(cp, original)

    def test_python_copy_warns_abt_adding_to_network(self, make_f, assert_f):
        original = make_f()
        copy(original)  # Fine because not in a network
        with nengo.Network():
            with pytest.warns(NotAddedToNetworkWarning):
                copy(original)


@pytest.mark.parametrize(
    "make_f", (make_ensemble, make_probe, make_node, make_connection, make_network)
)
class TestPickle:
    """A basic set of tests that should pass for all objects."""

    def test_pickle_roundtrip(self, make_f):
        """Tests making a deepcopy using pickling and unpickling"""
        original = make_f()
        cp = pickle.loads(pickle.dumps(original))
        assert_is_deepcopy(cp, original)

    def test_unpickling_warning_in_network(self, make_f):
        """Tests a not added to network warning when unpickling
        inside a network when the original was not in the network"""
        original = make_f()
        pkl = pickle.dumps(original)
        with nengo.Network():
            with pytest.warns(NotAddedToNetworkWarning):
                pickle.loads(pkl)


@pytest.mark.parametrize(
    "original",
    [
        nengo.learning_rules.PES(),
        nengo.learning_rules.BCM(),
        nengo.learning_rules.Oja(),
        nengo.learning_rules.Voja(),
        nengo.processes.WhiteNoise(),
        nengo.processes.FilteredNoise(),
        nengo.processes.BrownNoise(),
        nengo.processes.PresentInput([0.1, 0.2], 1.0),
        nengo.synapses.LinearFilter([0.1, 0.2], [0.3, 0.4], True),
        nengo.synapses.Lowpass(0.005),
        nengo.synapses.Alpha(0.005),
        nengo.synapses.Triangle(0.005),
    ],
)
class TestFrozenObjectCopies:
    def test_copy(self, original):
        assert_is_copy(copy(original), original)

    def test_pickle_roundtrip(self, original):
        assert_is_deepcopy(pickle.loads(pickle.dumps(original)), original)


def test_copy_spa(Simulator):
    with spa.SPA() as original:
        original.state = spa.State(16)
        original.cortex = spa.Cortical(spa.Actions("state = A"))

    cp = original.copy()

    # Check that it still builds.
    with Simulator(cp):
        pass

    # check vocab instance param is set
    for node in cp.all_nodes:
        if node.label in ["input", "output"]:
            assert cp.config[node].vocab is not None


def test_copy_instance_params():
    with nengo.Network() as orig_net:
        orig_net.config[nengo.Ensemble].set_param(
            "test", IntParam("test", optional=True)
        )
        orig_net.config[nengo.Ensemble].set_param(
            "test2", NdarrayParam("test2", optional=True)
        )

        orig_ens = nengo.Ensemble(10, 1)
        orig_net.config[orig_ens].test = 42
        orig_net.config[orig_ens].test2 = np.array([49])

    # test copy function
    copy_net = orig_net.copy()
    copy_ens = copy_net.ensembles[0]
    assert copy_net.config[copy_ens].test == 42
    assert np.array_equal(copy_net.config[copy_ens].test2, np.array([49]))

    # test copy via pickle
    copy_net = pickle.loads(pickle.dumps(orig_net))
    copy_ens = copy_net.ensembles[0]
    assert copy_net.config[copy_ens].test == 42
    assert np.array_equal(copy_net.config[copy_ens].test2, np.array([49]))


def test_pickle_model(Simulator, seed, allclose):
    t_run = 0.5
    simseed = seed + 1

    with nengo.Network(seed=seed) as network:
        u = nengo.Node(nengo.processes.WhiteSignal(t_run, 5))
        a = nengo.Ensemble(100, 1)
        b = nengo.Ensemble(100, 1)
        nengo.Connection(u, a, synapse=None)
        nengo.Connection(a, b, function=np.square)
        up = nengo.Probe(u, synapse=0.01)
        ap = nengo.Probe(a, synapse=0.01)
        bp = nengo.Probe(b, synapse=nengo.Alpha(0.01))

    with Simulator(network, seed=simseed) as sim:
        sim.run(t_run)
        t0, u0, a0, b0 = sim.trange(), sim.data[up], sim.data[ap], sim.data[bp]
        pkls = pickle.dumps(dict(model=sim.model, up=up, ap=ap, bp=bp))

    # reload model
    del network, sim, up, ap, bp
    pkl = pickle.loads(pkls)
    up, ap, bp = pkl["up"], pkl["ap"], pkl["bp"]

    with Simulator(None, model=pkl["model"], seed=simseed) as sim:
        sim.run(t_run)
        t1, u1, a1, b1 = sim.trange(), sim.data[up], sim.data[ap], sim.data[bp]

    tols = dict(atol=1e-5)
    assert allclose(t1, t0, **tols)
    assert allclose(u1, u0, **tols)
    assert allclose(a1, a0, **tols)
    assert allclose(b1, b0, **tols)


def test_copy_convolution():
    x = nengo.Convolution(1, (2, 3, 4), channels_last=False)
    y = copy(x)

    assert x.n_filters == y.n_filters
    assert x.input_shape == y.input_shape
    assert x.channels_last == y.channels_last


@pytest.mark.parametrize(
    "optimize, dt, progress_bar", [(True, 0.001, False), (False, 0.002, True)]
)
def test_pickle_sim(Simulator, seed, allclose, optimize, dt, progress_bar):
    trun0 = 0.5
    trun1 = 0.5
    simseed = seed + 1
    progress = TerminalProgressBar() if progress_bar else progress_bar

    with nengo.Network(seed=seed) as network:
        u = nengo.Node(nengo.processes.WhiteSignal(trun0 + trun1, high=5))
        a = nengo.Ensemble(100, 1)
        b = nengo.Ensemble(100, 1)
        nengo.Connection(u, a, synapse=None)
        nengo.Connection(a, b, function=np.square)
        up = nengo.Probe(u, synapse=0.01)
        ap = nengo.Probe(a, synapse=0.01)
        bp = nengo.Probe(b, synapse=nengo.Alpha(0.01))

    with Simulator(
        network, seed=simseed, dt=dt, optimize=optimize, progress_bar=progress
    ) as sim:
        sim.run(trun0)
        pkls = pickle.dumps(dict(sim=sim, up=up, ap=ap, bp=bp))

        sim.run(trun1)
        t0, u0, a0, b0 = sim.trange(), sim.data[up], sim.data[ap], sim.data[bp]

    # reload model
    del network, sim, up, ap, bp
    pkl = pickle.loads(pkls)
    up, ap, bp = pkl["up"], pkl["ap"], pkl["bp"]

    with pkl["sim"] as sim:
        assert sim.seed == simseed
        assert sim.dt == dt
        assert sim.optimize == optimize
        if progress_bar:
            assert isinstance(sim.progress_bar, TerminalProgressBar)
        else:
            assert sim.progress_bar is progress_bar

        sim.run(trun1)
        t1, u1, a1, b1 = sim.trange(), sim.data[up], sim.data[ap], sim.data[bp]

    tols = dict(atol=1e-5)
    assert allclose(t1, t0, **tols)
    assert allclose(u1, u0, **tols)
    assert allclose(a1, a0, **tols)
    assert allclose(b1, b0, **tols)

    # check that closed status is preserved
    pkl = pickle.loads(pickle.dumps(dict(sim=sim)))
    assert pkl["sim"].closed



"""Syntactic parsing of the subexpressions of all action expressions."""

from nengo.exceptions import SpaParseError
from nengo.utils.numpy import is_number


class Symbol:
    """A set of semantic pointer symbols and associated math.

    This is an abstract semantic pointer (not associated with a particular
    vocabulary or dimension). It is just meant for keeping track of the
    desired manipulations until such time as it is parsed with a particular
    `.Vocabulary`.

    Its contents are a single string, and this string is manipulated via
    standard mathematical operators (``+ - * ~``) for SemanticPointers.
    The result will always be able to be passed to `.Vocabulary.parse`
    to get a valid `.SemanticPointer`.

    This is used by the `.spa.Actions` parsing system.
    """

    def __init__(self, symbol):
        self.symbol = symbol

    def __add__(self, other):
        if is_number(other):
            other = Symbol("%g" % other)
        if isinstance(other, Symbol):
            return Symbol("(%s + %s)" % (self.symbol, other.symbol))
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Symbol):
            return Symbol("(%s - %s)" % (self.symbol, other.symbol))
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, Symbol):
            if other.symbol == "1":
                return self
            if self.symbol == "1":
                return other
            return Symbol("(%s * %s)" % (self.symbol, other.symbol))
        if is_number(other):
            if other == 1:
                return self
            if self.symbol == "1":
                return Symbol("%g" % other)
            return Symbol("(%s * %g)" % (self.symbol, other))
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __invert__(self):
        if self.symbol.startswith("~"):
            return Symbol(self.symbol[1:])
        else:
            return Symbol("~%s" % self.symbol)

    def __neg__(self):
        if self.symbol.startswith("-"):
            return Symbol(self.symbol[1:])
        else:
            return Symbol("-%s" % self.symbol)

    def __str__(self):
        return str(self.symbol)


class Source:
    """A particular source of a vector for the action system.

    This will always refer to a particular named output from a
    `.spa.module.Module`. It also tracks a single `.Symbol` which represents
    a desired transformation from that output. For example,
    ``Source('vision') * Symbol('VISION')`` will result in a ``Source`` object
    for ``'vision'``, but with transform set to the ``Symbol('VISION')``.

    This is used by the `.spa.Actions` parsing system.
    """

    def __init__(self, name, transform=Symbol("1"), inverted=False):
        self.name = name  # the name of the module output
        self.transform = transform  # the Symbol for the transformation
        self.inverted = inverted

    def __invert__(self):
        if self.transform.symbol != "1":
            raise SpaParseError("You can only invert sources without transforms")
        return Source(self.name, self.transform, not self.inverted)

    def __mul__(self, other):
        if isinstance(other, Source):
            return Convolution(self, other)
        elif is_number(other) or isinstance(other, Symbol):
            return Source(self.name, self.transform * other, self.inverted)
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return Source(self.name, -self.transform, self.inverted)

    def __add__(self, other):
        return Summation([self]).__add__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __str__(self):
        if self.transform.symbol == "1":
            trans_text = ""
        else:
            trans_text = "%s * " % self.transform
        if self.inverted:
            trans_text += "~"
        return "%s%s" % (trans_text, self.name)


class DotProduct:
    """The dot product of a Source and a Source or a Source and a Symbol.

    This represents a similarity measure for computing the utility of
    an action. It also maintains a scaling factor on the result,
    so that the 0.5 in ``"0.5*DotProduct(Source('vision'), Symbol('DOG'))"``
    can be correctly tracked.

    This is used by the `.spa.Actions` parsing system.
    """

    def __init__(self, item1, item2, scale=1.0):
        if isinstance(item1, (int, float)):
            item1 = Symbol(item1)
        if isinstance(item2, (int, float)):
            item2 = Symbol(item2)
        if not isinstance(item1, (Source, Symbol)):
            raise SpaParseError(
                "The first item in the dot product is not a "
                "semantic pointer or a spa.Module output."
            )
        if not isinstance(item2, (Source, Symbol)):
            raise SpaParseError(
                "The second item in the dot product is not a "
                "semantic pointer or a spa.Module output."
            )
        if not isinstance(item1, Source) and not isinstance(item2, Source):
            raise SpaParseError(
                "One of the two terms for the dot product "
                "must be a spa.Module output."
            )
        self.item1 = item1
        self.item2 = item2
        self.scale = float(scale)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return DotProduct(self.item1, self.item2, self.scale * other)
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        return self.__mul__(1.0 / other)

    def __truediv__(self, other):
        return self.__div__(other)

    def __neg__(self):
        return DotProduct(self.item1, self.item2, -self.scale)

    def __add__(self, other):
        return Summation([self]).__add__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __str__(self):
        if self.scale == 1:
            scale_text = ""
        elif self.scale == -1:
            scale_text = "-"
        else:
            scale_text = "%g * " % self.scale
        return "%sdot(%s, %s)" % (scale_text, self.item1, self.item2)


class Convolution:
    """The convolution of two sources together."""

    def __init__(self, source1, source2, transform=Symbol("1")):
        self.source1 = source1
        self.source2 = source2
        self.transform = transform

    def __mul__(self, other):
        if isinstance(other, (Symbol, int, float)):
            return Convolution(self.source1, self.source2, self.transform * other)
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return Convolution(self.source1, self.source2, -self.transform)

    def __add__(self, other):
        return Summation([self]).__add__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __str__(self):
        return "((%s) * (%s)) * %s" % (self.source1, self.source2, self.transform)


class Summation:
    """A summation over all subexpressions."""

    def __init__(self, items):
        self.items = items

    def __mul__(self, other):
        return Summation([x * other for x in self.items])

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        return self.__mul__(1.0 / other)

    def __truediv__(self, other):
        return self.__div__(other)

    def __add__(self, other):
        if isinstance(other, (int, float, Source, Symbol, DotProduct, Convolution)):
            return Summation(self.items + [other])
        if isinstance(other, Summation):
            return Summation(self.items + other.items)
        return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __neg__(self):
        return Summation([-x for x in self.items])

    def __str__(self):
        return " + ".join(str(v) for v in self.items)



import pytest

pytest_plugins = ["pytester"]


@pytest.mark.parametrize("xfail", (True, False))
def test_unsupported(xfail, testdir):
    """Test ``nengo_test_unsupported`` config option and ``--unsupported`` arg"""

    # Set up a dummy nengo package directory, so that `pytest_nengo.is_nengo_test`
    # returns True
    nengo_dir = testdir.mkpydir("nengo")

    # Create a test file with some dummy tests, and move it into nengo_dir
    test_file_path = testdir.makefile(
        ".py",
        test_file="""
        import pytest

        @pytest.mark.parametrize("param", (True, False))
        def test_unsupported(param):
            print(f"test_unsupported param={param} ran")
            assert param

        @pytest.mark.parametrize("param", (True, False))
        def test_unsupported_all(param):
            print(f"test_unsupported_all param={param} ran")
            assert False

        def test_supported():
            print("test_supported ran")
            assert True
        """,
    )
    test_file_path.move(test_file_path.new(dirname=nengo_dir))

    # Create the .ini file to skip/xfail the failing tests. This will
    # make sure square brackets for parameters just skip that parametrization.
    # We also make sure that both single-line and multiline comments work.
    testdir.makefile(
        ".ini",
        pytest="""
        [pytest]
        nengo_test_unsupported =
            test_file.py::test_unsupported[False]
                "One unsupported param
                with multiline comment"
            test_file.py::test_unsupported_all*
                "Two unsupported params with single-line comment"

        # avoid trying to load neurons from dummy nengo package
        nengo_neurons =
        """,
    )

    args = "-rsx -sv".split()
    if xfail:
        args.append("--unsupported")
    output = testdir.runpytest_subprocess(*args)

    # ensure that these lines appear somewhere in the output
    output.stdout.fnmatch_lines_random(
        [
            "*One unsupported param with multiline comment",
            "*Two unsupported params with single-line comment",
            "*test_supported ran",
            "*test_unsupported param=True ran",
        ]
    )

    # if `--unsupported`, unsupported tests run and xfail, otherwise they skip
    outcomes = output.parseoutcomes()
    if xfail:
        output.stdout.fnmatch_lines_random(
            [
                "*test_unsupported param=False ran",
                "*test_unsupported_all param=True ran",
                "*test_unsupported_all param=False ran",
            ]
        )
        assert outcomes["xfailed"] == 3
        assert "skipped" not in outcomes
    else:
        assert outcomes["skipped"] == 3
        assert "xfailed" not in outcomes
    assert "failed" not in outcomes
    assert outcomes["passed"] == 2


class MockSimulator:
    """A Simulator that does not support any tests."""


def test_pyargs(testdir):
    # mark all the tests as unsupported
    testdir.makefile(
        ".ini",
        pytest="""
        [pytest]
        nengo_simulator = nengo.tests.test_pytest_nengo.MockSimulator
        nengo_test_unsupported =
            *
                "Using mock simulator"
        """,
    )

    outcomes = testdir.runpytest_subprocess("--pyargs", "nengo").parseoutcomes()

    assert "failed" not in outcomes
    assert "passed" not in outcomes
    assert outcomes["skipped"] > 250
    assert outcomes["deselected"] > 700



import time

import numpy as np
import pytest

import nengo.utils.numpy as npext
from nengo.dists import (
    PDF,
    Choice,
    CosineSimilarity,
    DistOrArrayParam,
    Exponential,
    Gaussian,
    QuasirandomSequence,
    Samples,
    ScatteredHypersphere,
    SqrtBeta,
    Uniform,
    UniformHypersphere,
    _betaincinv22,
)
from nengo.exceptions import ValidationError


def test_pdf(rng, allclose):
    s = 0.25
    f = lambda x: (
        np.exp(-0.5 * (x + 0.5) ** 2 / s ** 2) + np.exp(-0.5 * (x - 0.5) ** 2 / s ** 2)
    )

    xref = np.linspace(-2, 2, 101)
    pref = f(xref)
    pref /= pref.sum()
    dist = PDF(xref, pref)

    n = 100000
    samples = dist.sample(n, rng=rng)
    h, xedges = np.histogram(samples, bins=101)
    x = 0.5 * (xedges[:-1] + xedges[1:])
    dx = np.diff(xedges)
    y = h / float(h.sum()) / dx
    z = f(x)
    z = z / z.sum() / dx
    assert allclose(y, z, atol=0.05)

    with pytest.raises(ValidationError, match="PDF must sum to one"):
        dist = PDF([0, 1, 2], [0.1, 1.1, 0.1])

    with pytest.raises(ValidationError, match="`x` and `p` must be the same length"):
        dist = PDF([0, 1], [0, 1, 0])


@pytest.mark.parametrize("low,high", [(-2, -1), (-4, 1), (1, 2), (1, -1)])
def test_uniform(low, high, rng, allclose):
    n = 200
    dist = Uniform(low, high)
    samples = dist.sample(n, rng=rng)
    if low < high:
        assert np.all(samples >= low)
        assert np.all(samples < high)
    else:
        assert np.all(samples <= low)
        assert np.all(samples > high)
    histogram, _ = np.histogram(samples, bins=5)
    assert allclose(histogram, np.mean(histogram), atol=0.1 * n)

    # test `integer=true`
    dist = Uniform(low, high, integer=True)
    if low < high:
        samples = dist.sample(n, rng=rng)
        assert np.all(samples >= low)
        assert np.all(samples < high)
        assert np.all(samples % 1 == 0)
    else:
        with pytest.raises(ValueError):
            samples = dist.sample(n, rng=rng)


@pytest.mark.parametrize("mean,std", [(0, 1), (0, 0), (10, 2)])
def test_gaussian(mean, std, rng):
    n = 500
    if std <= 0:
        with pytest.raises(ValueError):
            dist = Gaussian(mean, std)
    else:
        dist = Gaussian(mean, std)
        samples = dist.sample(n, rng=rng)
        assert abs(np.mean(samples) - mean) < 3 * std / np.sqrt(n)
        assert abs(np.std(samples) - std) < 0.25


@pytest.mark.parametrize(
    "scale,shift,high", [(1.0, 0.0, np.inf), (10.0, 0.0, 1.0), (0.1, 0.3, 1.0)]
)
def test_exponential(scale, shift, high, rng):
    n = 100
    dist = Exponential(scale, shift=shift, high=high)
    samples = dist.sample(n, rng=rng)
    assert np.all(samples >= shift)
    assert np.all(samples <= high)
    # approximation of 95% confidence interval
    ci = scale * 1.96 / np.sqrt(n)
    if scale + ci < high:
        assert abs(np.mean(samples - shift) - scale) < ci


@pytest.mark.parametrize("cls", [UniformHypersphere, ScatteredHypersphere])
@pytest.mark.parametrize(
    "min_magnitude,d", [(0, 1), (0, 2), (0, 5), (0.6, 1), (0.3, 2), (0.4, 5)]
)
def test_hypersphere_volume(cls, min_magnitude, d, rng, allclose):
    n = 250 * d
    dist = cls(min_magnitude=min_magnitude)
    samples = dist.sample(n, d, rng=rng)
    assert samples.shape == (n, d)
    assert allclose(np.mean(samples, axis=0), 0, atol=0.1)

    norms = npext.norm(samples, axis=1)
    assert np.all(norms >= min_magnitude)
    assert np.all(norms <= 1)

    # probability of not finding a point in [min_magnitude, r_tol_min], [r_tol_max, 1]
    q = 1e-5
    r_min_d = min_magnitude ** d
    r_tol_min = (r_min_d + (1 - r_min_d) * (1 - q ** (1 / n))) ** (1 / d)
    assert norms.min() <= r_tol_min
    r_tol_max = (1 - (1 - r_min_d) * (1 - q ** (1 / n))) ** (1 / d)
    assert norms.max() >= r_tol_max


@pytest.mark.parametrize("cls", [UniformHypersphere, ScatteredHypersphere])
@pytest.mark.parametrize("dimensions", [1, 2, 5])
def test_hypersphere_surface(cls, dimensions, rng, allclose):
    n = 200 * dimensions
    dist = cls(surface=True)
    samples = dist.sample(n, dimensions, rng=rng)
    assert samples.shape == (n, dimensions)
    assert allclose(npext.norm(samples, axis=1), 1)
    assert allclose(np.mean(samples, axis=0), 0, atol=0.25 / dimensions)


@pytest.mark.parametrize("cls", [UniformHypersphere, ScatteredHypersphere])
def test_hypersphere_errors(cls):
    with pytest.raises(ValidationError, match="Must be of type 'bool'"):
        cls(surface=0)

    with pytest.raises(ValidationError, match="Dimensions must be a positive integer"):
        cls().sample(1, d=-1)


@pytest.mark.parametrize("cls", [UniformHypersphere, ScatteredHypersphere])
def test_hypersphere_warns(cls):
    with pytest.warns(UserWarning, match="min_magnitude ignored because surface"):
        cls(surface=True, min_magnitude=0.1)


def test_quasirandom_sequence_phi():
    def phi(x, iters=100):
        y = 1
        for _ in range(iters):
            y = (1 + y) ** (1 / (x + 1))
        return y

    rd = QuasirandomSequence()
    for i in range(1, 20):
        assert np.allclose(rd._phi(i), phi(i)), str(i)

    with pytest.raises(RuntimeError, match="did not converge"):
        print(rd._phi(np.nan))


@pytest.mark.parametrize("dims", [2, 3, 7, 8])
@pytest.mark.parametrize("surface", [True, False])
def test_scattered_hypersphere(dims, surface, seed, plt):
    scipy_special = pytest.importorskip("scipy.special")

    n = 3000
    dists = [
        UniformHypersphere(surface=surface),
        ScatteredHypersphere(surface=surface, method="sct"),
        ScatteredHypersphere(surface=surface, method="sct-approx"),
        ScatteredHypersphere(surface=surface, method="tfww"),
    ]
    assert isinstance(dists[0], UniformHypersphere)

    xx = []  # generated points, for each dist
    times = []  # time taken to generate the points, for each dist
    for dist in dists:
        rng = np.random.RandomState(seed)
        timer = time.time()
        x = dist.sample(n, d=dims, rng=rng)
        timer = time.time() - timer
        rng.shuffle(x)  # shuffle so we can compute distances in blocks without bias
        xx.append(x)
        times.append(timer)

    dd = []  # distance to the nearest point for each point, for each dist
    rr = []  # radii (norms) of all the generated points, for each dist
    for x in xx:
        # compute distances in blocks for efficiency (this means we're not actually
        # getting the minimum distance, just a proxy)
        n_split = 1000
        d_min = []
        for i in range(0, n, n_split):
            xi = x[i : i + n_split]
            d2 = ((xi[:, :, None] - xi.T[None, :, :]) ** 2).sum(axis=1)
            np.fill_diagonal(d2, np.inf)
            d_min.append(np.sqrt(d2.min(axis=1)))
        d_min = np.concatenate(d_min)
        dd.append(d_min)
        rr.append(np.sqrt((x ** 2).sum(axis=1)))

    # compute the approximate distance between points if they were evenly spread
    volume = np.pi ** (0.5 * dims) / scipy_special.gamma(0.5 * dims + 1)
    if surface:
        volume *= dims
    even_distance = (volume / n) ** (1 / (dims - 1 if surface else dims))

    # --- plots
    colors = ["b", "g", "r", "m", "c"]

    plt.subplot(211)
    bins = np.linspace(np.min(dd), np.max(dd), 31)
    for i, d in enumerate(dd):
        histogram, _ = np.histogram(d, bins=bins)
        plt.plot(
            0.5 * (bins[:-1] + bins[1:]),
            histogram,
            colors[i],
        )
        plt.plot([d.min()], [0], colors[i] + "x")
    plt.plot([even_distance], [0], "kx")
    plt.title(f"surface={surface}, dims={dims}, n={n}")

    plt.subplot(212)
    bins = np.linspace(0, 1.1, 31)
    for i, r in enumerate(rr):
        histogram, _ = np.histogram(r, bins=bins)
        plt.plot(
            0.5 * (bins[:-1] + bins[1:]),
            histogram,
            colors[i],
            label=f"{dists[i]}: t={times[i]:0.2e}",
        )
    plt.legend()

    # --- checks
    uniform_min = dd[0].min()
    for i, dist in enumerate(dists):
        if i == 0:
            continue

        # check that we're significantly better than UniformHypersphere
        d_min = dd[i].min()
        assert d_min > 1.2 * uniform_min, str(dist)

        # check that all surface points are on the surface
        if surface:
            assert np.allclose(rr[i], 1.0, atol=1e-5), str(dist)


@pytest.mark.parametrize("weights", [None, [5, 1, 2, 9], [3, 2, 1, 0]])
def test_choice(weights, rng, allclose):
    """Tests the choice function with weights"""
    n = 1000
    choices = [[1, 1], [1, -1], [-1, 1], [-1, -1]]
    N = len(choices)

    dist = Choice(choices, weights=weights)
    # If d is passed, it has to match
    with pytest.raises(ValueError):
        dist.sample(n, d=4, rng=rng)
    sample = dist.sample(n, rng=rng)
    tsample = [tuple(point) for point in sample]
    tchoices = [tuple(choice) for choice in choices]

    # check that frequency of choices matches weights
    inds = [tchoices.index(s) for s in tsample]
    histogram, _ = np.histogram(inds, bins=np.linspace(-0.5, N - 0.5, N + 1))
    p_empirical = histogram / float(histogram.sum())
    p = np.ones(N) / N if dist.p is None else dist.p
    sterr = 1.0 / np.sqrt(n)  # expected maximum standard error
    assert allclose(p, p_empirical, atol=2 * sterr)


def test_choice_errors():
    with pytest.raises(ValidationError, match="Number of weights.*must match.*options"):
        Choice([2], [1, 2, 3])

    with pytest.raises(ValidationError, match="All weights must be non-negative"):
        Choice([2], [-1])

    with pytest.raises(ValidationError, match="Sum of weights must be positive"):
        Choice([1, 2], [0, 0])

    with pytest.raises(ValidationError, match="Options must be of dimensionality 1"):
        Choice([0]).sample(n=2, d=1)


@pytest.mark.parametrize("shape", [(12, 2), (7, 1), (7,), (1, 1)])
def test_samples(shape, rng, allclose):
    samples = rng.random_sample(size=shape)
    d = Samples(samples)
    dims = None if len(shape) == 1 else shape[1]
    assert allclose(d.sample(shape[0], dims), samples)


@pytest.mark.parametrize("samples", [[1.0, 2.0, 3.0], [[1, 2], [3, 4]]])
def test_samples_list(samples, allclose):
    d = Samples(samples)
    shape = np.array(samples).shape
    dims = None if len(shape) == 1 else shape[1]
    assert allclose(d.sample(shape[0], dims), samples)


def test_samples_errors(rng):
    samples = rng.random_sample(size=(12, 2))
    with pytest.raises(ValidationError):
        Samples(samples).sample(11, 2)
    with pytest.raises(ValidationError):
        Samples(samples).sample(12, 1)
    with pytest.raises(ValidationError):
        Samples(samples).sample(12)

    samples = rng.random_sample(size=12)
    with pytest.raises(ValidationError):
        Samples(samples).sample(12, 2)


@pytest.mark.parametrize("n,m", [(99, 1), (50, 50)])
def test_sqrt_beta(n, m, rng):
    num_samples = 1000
    num_bins = 5

    vectors = rng.randn(num_samples, n + m)
    vectors /= npext.norm(vectors, axis=1, keepdims=True)
    expectation, _ = np.histogram(npext.norm(vectors[:, :m], axis=1), bins=num_bins)

    dist = SqrtBeta(n, m)
    samples = dist.sample(num_samples, 1, rng=rng)
    histogram, _ = np.histogram(samples, bins=num_bins)

    assert np.all(np.abs(np.asfarray(histogram - expectation) / num_samples) < 0.16)


@pytest.mark.parametrize("n,m", [(4, 1), (10, 5)])
def test_sqrt_beta_analytical(n, m, rng, allclose):
    """Tests pdf, cdf, and ppf of SqrtBeta distribution."""
    pytest.importorskip("scipy")  # beta and betainc

    dt = 0.001
    x = np.arange(dt, 1 + dt, dt)

    dist = SqrtBeta(n, m)

    pdf = dist.pdf(x)
    cdf = dist.cdf(x)
    ppf = dist.ppf(cdf)

    # The pdf should reflect the samples
    num_samples = 2500
    num_bins = 5

    samples = dist.sample(num_samples, rng=rng)
    act_hist, _ = np.histogram(samples, bins=num_bins)
    bin_points = np.linspace(0, 1, num_bins + 1)
    bin_cdf = dist.cdf(bin_points)
    exp_freq = bin_cdf[1:] - bin_cdf[:-1]
    assert np.all(np.abs(np.asfarray(act_hist) / num_samples - exp_freq) < 0.1)

    # The cdf should be the accumulated pdf
    assert allclose(cdf, np.cumsum(pdf) * dt, atol=0.01)

    # The ppf should give back x
    assert allclose(x, ppf, atol=0.01)


@pytest.mark.parametrize("d", [2, 3, 10, 50])
def test_cosine_similarity(d, rng):
    """Tests CosineSimilarity sampling."""
    num_samples = 2500
    num_bins = 8

    # Check that it gives a single dimension from UniformHypersphere
    exp_dist = UniformHypersphere(surface=True)
    act_dist = CosineSimilarity(d)

    exp = exp_dist.sample(num_samples, d, rng=rng)[:, 0]
    act = act_dist.sample(num_samples, rng=rng)

    exp_hist, _ = np.histogram(exp, bins=num_bins)
    act_hist, _ = np.histogram(act, bins=num_bins)

    assert np.all(np.abs(np.asfarray(exp_hist - act_hist) / num_samples) < 0.15)


@pytest.mark.parametrize("d", [2, 3, 10])
def test_cosine_analytical(d, allclose):
    pytest.importorskip("scipy")  # beta, betainc, betaincinv

    dt = 0.0001
    x = np.arange(-1 + dt, 1, dt)

    def p(x, d):
        # unnormalized CosineSimilarity distribution, derived by Eric H.
        return (1 - x * x) ** ((d - 3) / 2.0)

    dist = CosineSimilarity(d)

    pdf_exp = dist.pdf(x)
    pdf_act = p(x, d)

    cdf_exp = dist.cdf(x)
    cdf_act = np.cumsum(pdf_act) / np.sum(pdf_act)

    # Check that we get the expected pdf after normalization
    assert allclose(pdf_exp / np.sum(pdf_exp), pdf_act / np.sum(pdf_act), atol=0.01)

    # Check that this accumulates to the expected cdf
    assert allclose(cdf_exp, cdf_act, atol=0.01)

    # Check that the inverse cdf gives back x
    assert allclose(dist.ppf(cdf_exp), x, atol=0.01)


def test_cosine_sample_shape(seed, allclose):
    """Tests that CosineSimilarity sample has correct shape."""
    # sampling (n, d) should be the exact same as sampling (n*d,)
    n = 3
    d = 4
    dist = CosineSimilarity(2)
    a = dist.sample(n, d, rng=np.random.RandomState(seed))
    b = dist.sample(n * d, rng=np.random.RandomState(seed))
    assert allclose(a.flatten(), b)


@pytest.mark.parametrize("d,p", [(3, 0), (5, 0.4), (10, 0.7), (50, 1.0)])
def test_cosine_intercept(d, p, rng, allclose):
    """Tests CosineSimilarity inverse cdf for finding intercepts."""
    pytest.importorskip("scipy")  # betaincinv

    num_samples = 500

    exp_dist = UniformHypersphere(surface=True)
    act_dist = CosineSimilarity(d)

    dots = exp_dist.sample(num_samples, d, rng=rng)[:, 0]

    # Find the desired intercept so that dots >= c with probability p
    c = act_dist.ppf(1 - p)
    assert allclose(np.sum(dots >= c) / float(num_samples), p, atol=0.05)


def test_distorarrayparam():
    """DistOrArrayParams can be distributions or samples."""

    class Test:
        dp = DistOrArrayParam("dp", default=None, sample_shape=["*", "*"])

    inst = Test()
    inst.dp = UniformHypersphere()
    assert isinstance(inst.dp, UniformHypersphere)
    inst.dp = np.array([[1], [2], [3]])
    assert np.all(inst.dp == np.array([[1], [2], [3]]))
    with pytest.raises(ValueError):
        inst.dp = "a"
    # Sample must have correct dims
    with pytest.raises(ValueError):
        inst.dp = np.array([1])


def test_distorarrayparam_sample_shape():
    """sample_shape dictates the shape of the sample that can be set."""

    class Test:
        dp = DistOrArrayParam("dp", default=None, sample_shape=["d1", 10])
        d1 = 4

    inst = Test()
    # Distributions are still cool
    inst.dp = UniformHypersphere()
    assert isinstance(inst.dp, UniformHypersphere)
    # Must be shape (4, 10)
    inst.dp = np.ones((4, 10))
    assert np.all(inst.dp == np.ones((4, 10)))
    with pytest.raises(ValidationError):
        inst.dp = np.ones((10, 4))
    assert np.all(inst.dp == np.ones((4, 10)))


def test_frozen():
    """Test attributes inherited from FrozenObject"""
    a = Uniform(-0.3, 0.6)
    b = Uniform(-0.3, 0.6)
    c = Uniform(-0.2, 0.6)

    assert hash(a) == hash(a)
    assert hash(b) == hash(b)
    assert hash(c) == hash(c)

    assert a == b
    assert hash(a) == hash(b)
    assert a != c
    assert hash(a) != hash(c)  # not guaranteed, but highly likely
    assert b != c
    assert hash(b) != hash(c)  # not guaranteed, but highly likely


def _test_betaincinv22(plt, allclose):
    scipy_special = pytest.importorskip("scipy.special")

    # call once to load table, so that doesn't effect timing
    _betaincinv22.lookup(5, [0.1])

    dims = np.concatenate(
        [np.arange(1, 50), np.round(np.logspace(np.log10(51), 3.1)).astype(np.int64)]
    )
    x = np.linspace(0, 1, 1000)

    results = []
    for dim in dims:
        ref_timer = time.time()
        yref = scipy_special.betaincinv(dim / 2, 0.5, x)
        ref_timer = time.time() - ref_timer

        timer = time.time()
        y = _betaincinv22.lookup(dim, x)
        timer = time.time() - timer

        results.append((yref, y, ref_timer, timer))

    n_show = 5
    resultsT = list(zip(*results))
    errors = np.abs(np.array(resultsT[0]) - np.array(resultsT[1])).max(axis=1)
    show_inds = np.argsort(errors)[-n_show:]

    subplots = plt.subplots(nrows=2, sharex=True)
    if isinstance(subplots, tuple):
        _, ax = subplots

        for i in show_inds:
            yref, y, ref_timer, timer = results[i]
            dim = dims[i]

            ax[0].plot(x, y, label=f"dims={dim}")
            ax[1].plot(x, y - yref)

        speedups = np.array(resultsT[2]) / np.array(resultsT[3])
        ax[0].set_title(f"average speedup = {speedups.mean():0.1f} times")
        ax[0].set_ylabel("value")
        ax[1].set_xlabel("input")
        ax[1].set_ylabel("error")
        ax[0].legend()

    for i, (yref, y, ref_timer, timer) in enumerate(results):
        # allow error to increase for higher dimensions (to 5e-3 when dims=1000)
        atol = 1e-3 + (np.log10(dims[i]) / 3) * 4e-3
        assert allclose(y, yref, atol=atol), f"dims={dims[i]}"


def test_make_betaincinv22_table(monkeypatch, tmp_path, plt, allclose):
    pytest.importorskip("scipy.special")
    test_path = str(tmp_path / "betaincinv22_test_table.npz")
    monkeypatch.setattr(_betaincinv22, "path", test_path)
    monkeypatch.setattr(_betaincinv22, "table", None)

    _betaincinv22.make_table(n_interp=200, n_dims=50)
    _test_betaincinv22(plt=plt, allclose=allclose)


def test_betaincinv22_lookup(monkeypatch, plt, allclose):
    pytest.importorskip("scipy.special")
    monkeypatch.setattr(_betaincinv22, "table", None)
    _test_betaincinv22(plt=plt, allclose=allclose)


def test_betaincinv22_errors():
    x = np.linspace(0.1, 0.9)
    _betaincinv22.lookup(3, x)

    with pytest.raises(ValidationError, match="must be an integer >= 1"):
        _betaincinv22.lookup(0, x)

    with pytest.raises(ValidationError, match="must be an integer >= 1"):
        _betaincinv22.lookup(2.2, x)



import numpy as np
import pytest

import nengo
from nengo.exceptions import ValidationError
from nengo.processes import WhiteSignal
from nengo.synapses import Alpha, LinearFilter, Lowpass, Synapse, SynapseParam, Triangle
from nengo.utils.filter_design import cont2discrete
from nengo.utils.testing import signals_allclose

# The following num, den are for a 4th order analog Butterworth filter,
# generated with `scipy.signal.butter(4, 0.1, analog=False)`
butter_num = np.array([0.0004166, 0.0016664, 0.0024996, 0.0016664, 0.0004166])
butter_den = np.array([1.0, -3.18063855, 3.86119435, -2.11215536, 0.43826514])


def run_synapse(
    Simulator, seed, synapse, dt=1e-3, runtime=0.2, high=100, n_neurons=None
):
    model = nengo.Network(seed=seed)
    with model:
        u = nengo.Node(output=WhiteSignal(runtime, high=high))

        if n_neurons is not None:
            a = nengo.Ensemble(n_neurons, 1)
            nengo.Connection(u, a, synapse=None)
            target = a
        else:
            target = u

        ref = nengo.Probe(target)
        filtered = nengo.Probe(target, synapse=synapse)

    with Simulator(model, dt=dt, seed=seed + 1) as sim:
        sim.run(runtime)

    return sim.trange(), sim.data[ref], sim.data[filtered]


def test_direct(Simulator, plt, seed, allclose):
    dt = 1e-3
    a = 0.7

    synapse = LinearFilter([a], [1], analog=False)
    t, x, yhat = run_synapse(Simulator, seed, synapse, dt=dt)
    y = synapse.filt(x, dt=dt, y0=0)

    assert signals_allclose(t, y, yhat, delay=dt, allclose=allclose)
    assert signals_allclose(t, a * x, y, plt=plt, allclose=allclose)


def test_lowpass(Simulator, plt, seed, allclose):
    dt = 1e-3
    tau = 0.03

    t, x, yhat = run_synapse(Simulator, seed, Lowpass(tau), dt=dt)
    y = Lowpass(tau).filt(x, dt=dt, y0=0)

    assert signals_allclose(t, y, yhat, delay=dt, plt=plt, allclose=allclose)


def test_alpha(Simulator, plt, seed, allclose):
    dt = 1e-3
    tau = 0.03
    num, den = [1], [tau ** 2, 2 * tau, 1]

    t, x, yhat = run_synapse(Simulator, seed, Alpha(tau), dt=dt)
    y = LinearFilter(num, den).filt(x, dt=dt, y0=0)

    assert signals_allclose(t, y, yhat, delay=dt, atol=5e-6, plt=plt, allclose=allclose)


def test_triangle(Simulator, plt, seed, allclose):
    dt = 1e-3
    tau = 0.03

    t, x, ysim = run_synapse(Simulator, seed, Triangle(tau), dt=dt)
    yfilt = Triangle(tau).filt(x, dt=dt, y0=0)

    # compare with convolved filter
    n_taps = int(round(tau / dt)) + 1
    num = np.arange(n_taps, 0, -1, dtype=nengo.rc.float_dtype)
    num /= num.sum()
    y = np.convolve(x.ravel(), num)[: len(t)]
    y.shape = (-1, 1)

    assert allclose(y, yfilt, rtol=0)
    assert signals_allclose(t, y, ysim, delay=dt, rtol=0, plt=plt, allclose=allclose)

    # test y0 != 0
    assert allclose(Triangle(tau).filt(np.ones(100), dt=dt, y0=1), 1)


def test_decoders(Simulator, plt, seed, allclose):
    dt = 1e-3
    tau = 0.01

    t, x, yhat = run_synapse(Simulator, seed, Lowpass(tau), dt=dt, n_neurons=100)

    y = Lowpass(tau).filt(x, dt=dt, y0=0)
    assert signals_allclose(t, y, yhat, delay=dt, plt=plt, allclose=allclose)


def test_linearfilter(Simulator, plt, seed, allclose):
    dt = 1e-3
    synapse = LinearFilter(butter_num, butter_den, analog=False)
    t, x, yhat = run_synapse(Simulator, seed, synapse, dt=dt)
    y = synapse.filt(x, dt=dt, y0=0)

    assert signals_allclose(t, y, yhat, delay=dt, plt=plt, allclose=allclose)


def test_linearfilter_evaluate(plt):
    tau = 0.02
    ord1 = LinearFilter([1], [tau, 1])
    ord2 = LinearFilter([1], [tau ** 2, 2 * tau, 1])

    f = np.logspace(-1, 3, 100)
    y1 = ord1.evaluate(f)
    y2 = ord2.evaluate(f)

    plt.subplot(211)
    plt.semilogx(f, 20 * np.log10(np.abs(y1)))
    plt.semilogx(f, 20 * np.log10(np.abs(y2)))

    plt.subplot(212)
    plt.semilogx(f, np.angle(y1))
    plt.semilogx(f, np.angle(y2))

    jw_tau = 2.0j * np.pi * f * tau
    y1_ref = 1 / (jw_tau + 1)
    y2_ref = 1 / (jw_tau ** 2 + 2 * jw_tau + 1)
    assert np.allclose(y1, y1_ref)
    assert np.allclose(y2, y2_ref)


def test_linearfilter_y0(allclose):
    # --- y0 sets initial state correctly for high-order filter
    synapse = LinearFilter(butter_num, butter_den, analog=False)
    v = 9.81
    x = v * np.ones(10)
    assert allclose(synapse.filt(x, y0=v), v)
    assert not allclose(synapse.filt(x, y0=0), v, record_rmse=False, print_fail=0)

    # --- y0 does not work for high-order synapse when DC gain is zero
    synapse = LinearFilter([1, 0], [1, 1])
    with pytest.raises(ValidationError, match="Cannot solve for state"):
        synapse.filt(np.ones(10), y0=1)


def test_linearfilter_extras(allclose):
    # This filter is just a gain, but caused index errors previously
    synapse = nengo.LinearFilter([3], [2])
    assert allclose(synapse.filt([2.0]), 3)

    # differentiator should work properly
    diff = nengo.LinearFilter([1, -1], [1, 0], analog=False)
    assert allclose(diff.filt([1.0, -1.0, 2.0], y0=0), [1.0, -2.0, 3.0])

    # Filtering an integer array should cast to a float
    x = np.arange(10, dtype=nengo.rc.int_dtype)
    synapse = nengo.LinearFilter([1], [0.005, 1])
    assert synapse.filt(x).dtype == nengo.rc.float_dtype

    # Throw an error if non-float dtype
    shape = (1,)
    with pytest.raises(ValidationError, match="Only float data types"):
        synapse.make_state(shape, shape, dt=0.001, dtype=np.int32)
    with pytest.raises(ValidationError, match="Only float data types"):
        synapse.make_state(shape, shape, dt=0.001, dtype=np.complex64)


def test_step_errors():
    # error for A.shape[0] != B.shape[0]
    A = np.ones((2, 2))
    B = np.ones((1, 1))
    C = np.ones((1, 2))
    D = np.ones((1, 1))
    X = np.ones((2, 10))
    with pytest.raises(ValidationError, match="Matrices do not meet"):
        LinearFilter.General(A, B, C, D, X)


def test_filt(plt, rng, allclose):
    dt = 1e-3
    tend = 0.5
    t = dt * np.arange(tend / dt)
    nt = len(t)

    tau = 0.1
    tau_dt = tau / dt

    u = rng.normal(size=nt)

    tk = np.arange(0, 30 * tau_dt)
    k = 1.0 / tau_dt * np.exp(-tk / tau_dt)
    x = np.convolve(u, k, mode="full")[:nt]

    # support lists as input
    y = Lowpass(tau).filt(list(u), dt=dt, y0=0)

    plt.plot(t, x)
    plt.plot(t, y)

    assert allclose(x, y, atol=1e-3, rtol=1e-2)


def test_filtfilt(plt, rng, allclose):
    dt = 1e-3
    tend = 3.0
    t = dt * np.arange(tend / dt)
    nt = len(t)
    synapse = Lowpass(0.03)

    u = rng.normal(size=nt)
    x = synapse.filt(u, dt=dt)
    x = synapse.filt(x[::-1], y0=x[-1], dt=dt)[::-1]
    y = synapse.filtfilt(u, dt=dt)

    plt.plot(t, x)
    plt.plot(t, y, "--")

    assert allclose(x, y)


def test_lti_lowpass(rng, plt, allclose):
    dt = 1e-3
    tend = 3.0
    t = dt * np.arange(tend / dt)
    nt = len(t)

    tau = 1e-2
    lti = LinearFilter([1], [tau, 1])

    u = rng.normal(size=(nt, 10))
    x = Lowpass(tau).filt(u, dt=dt)
    y = lti.filt(u, dt=dt)

    plt.plot(t, x[:, 0], label="Lowpass")
    plt.plot(t, y[:, 0], label="LTI")
    plt.legend(loc="best")

    assert allclose(x, y)


def test_linearfilter_combine(rng, allclose):
    nt = 3000
    tau0, tau1 = 0.01, 0.02
    u = rng.normal(size=(nt, 10))
    x = LinearFilter([1], [tau0 * tau1, tau0 + tau1, 1]).filt(u, y0=0)
    y = Lowpass(tau0).combine(Lowpass(tau1)).filt(u, y0=0)
    assert allclose(x, y)

    with pytest.raises(ValidationError, match="other LinearFilters"):
        Lowpass(0.1).combine(Triangle(0.01))

    with pytest.raises(ValidationError, match="analog and digital"):
        Lowpass(0.1).combine(LinearFilter([1], [1], analog=False))


def test_combined_delay(Simulator, allclose):
    # ensure that two sequential filters has the same output
    # as combining their individual discrete transfer functions
    nt = 50
    tau = 0.005
    dt = 0.001

    sys1 = nengo.Lowpass(tau)
    (num,), den, _ = cont2discrete((sys1.num, sys1.den), dt=dt)
    sys2 = nengo.LinearFilter(np.poly1d(num) ** 2, np.poly1d(den) ** 2, analog=False)

    with nengo.Network() as model:
        u = nengo.Node(1)
        x = nengo.Node(size_in=1)
        nengo.Connection(u, x, synapse=sys1)
        p1 = nengo.Probe(x, synapse=sys1)
        p2 = nengo.Probe(u, synapse=sys2)

    with Simulator(model, dt=dt) as sim:
        sim.run_steps(nt)

    assert allclose(sim.data[p1], sim.data[p2])

    # Both have two time-step delays:
    # for sys1, this comes from two levels of filtering
    # for sys2, this comes from one level of filtering + delay in sys2
    assert allclose(sim.data[p1][:2], 0)
    assert not allclose(sim.data[p1][2], 0, record_rmse=False, print_fail=0)


def test_synapseparam():
    """SynapseParam must be a Synapse, and converts numbers to Lowpass."""

    class Test:
        sp = SynapseParam("sp", default=Lowpass(0.1))

    inst = Test()
    assert isinstance(inst.sp, Lowpass)
    assert inst.sp.tau == 0.1
    # Numbers are converted to Lowpass
    inst.sp = 0.05
    assert isinstance(inst.sp, Lowpass)
    assert inst.sp.tau == 0.05
    # None has meaning
    inst.sp = None
    assert inst.sp is None
    # Non-synapse not OK
    with pytest.raises(ValueError):
        inst.sp = "a"


def test_frozen():
    """Test attributes inherited from FrozenObject"""
    a = LinearFilter([1], [0.04, 1])
    b = LinearFilter([1], [0.04, 1])
    c = LinearFilter([1], [0.04, 1.1])

    assert hash(a) == hash(a)
    assert hash(b) == hash(b)
    assert hash(c) == hash(c)

    assert a == b
    assert hash(a) == hash(b)
    assert a != c
    assert hash(a) != hash(c)  # not guaranteed, but highly likely
    assert b != c
    assert hash(b) != hash(c)  # not guaranteed, but highly likely

    with pytest.raises((ValueError, RuntimeError)):
        a.den[0] = 9


def test_synapse_subclass(Simulator):
    class MySynapse(Synapse):  # pylint: disable=abstract-method
        pass

    with nengo.Network() as net:
        node_a = nengo.Node([0])
        node_b = nengo.Node(size_in=1)
        nengo.Connection(node_a, node_b, synapse=MySynapse())

    with pytest.raises(NotImplementedError, match="must implement make_state"):
        with Simulator(net):
            pass



import numpy as np

import nengo
from nengo.spa import Input, Vocabulary
from nengo.spa.assoc_mem import AssociativeMemory
from nengo.spa.utils import similarity


def test_am_spa_interaction(Simulator, seed, rng):
    """Make sure associative memory interacts with other SPA modules."""
    D = 16
    vocab = Vocabulary(D, rng=rng)
    vocab.parse("A+B+C+D")

    D2 = int(D / 2)
    vocab2 = Vocabulary(D2, rng=rng)
    vocab2.parse("A+B+C+D")

    def input_func(t):
        return "0.49*A" if t < 0.5 else "0.79*A"

    with nengo.spa.SPA(seed=seed) as m:
        m.buf = nengo.spa.Buffer(D)
        m.input = nengo.spa.Input(buf=input_func)

        m.am = AssociativeMemory(
            vocab,
            vocab2,
            input_keys=["A", "B", "C"],
            output_keys=["B", "C", "D"],
            default_output_key="A",
            threshold=0.5,
            inhibitable=True,
            wta_output=True,
            threshold_output=True,
        )

        cortical_actions = nengo.spa.Actions("am = buf")
        m.c_act = nengo.spa.Cortical(cortical_actions)

    # Check to see if model builds properly. No functionality test needed
    with Simulator(m):
        pass


def test_am_spa_keys_as_expressions(Simulator, plt, seed, rng):
    """Provide semantic pointer expressions as input and output keys."""
    D = 64

    vocab_in = Vocabulary(D, rng=rng)
    vocab_out = Vocabulary(D, rng=rng)

    vocab_in.parse("A+B")
    vocab_out.parse("C+D")

    in_keys = ["A", "A*B"]
    out_keys = ["C*D", "C+D"]

    with nengo.spa.SPA(seed=seed) as model:
        model.am = AssociativeMemory(
            input_vocab=vocab_in,
            output_vocab=vocab_out,
            input_keys=in_keys,
            output_keys=out_keys,
        )

        model.inp = Input(am=lambda t: "A" if t < 0.1 else "A*B")

        in_p = nengo.Probe(model.am.input)
        out_p = nengo.Probe(model.am.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    # Specify t ranges
    t = sim.trange()
    t_item1 = (t > 0.075) & (t < 0.1)
    t_item2 = (t > 0.175) & (t < 0.2)

    # Modify vocabularies (for plotting purposes)
    vocab_in.add(in_keys[1], vocab_in.parse(in_keys[1]).v)
    vocab_out.add(out_keys[0], vocab_out.parse(out_keys[0]).v)

    plt.subplot(2, 1, 1)
    plt.plot(t, similarity(sim.data[in_p], vocab_in))
    plt.ylabel("Input: " + ", ".join(in_keys))
    plt.legend(vocab_in.keys, loc="best")
    plt.ylim(top=1.1)
    plt.subplot(2, 1, 2)
    plt.plot(t, similarity(sim.data[out_p], vocab_out))
    plt.plot(t[t_item1], np.ones(t.shape)[t_item1] * 0.9, c="r", lw=2)
    plt.plot(t[t_item2], np.ones(t.shape)[t_item2] * 0.91, c="g", lw=2)
    plt.plot(t[t_item2], np.ones(t.shape)[t_item2] * 0.89, c="b", lw=2)
    plt.ylabel("Output: " + ", ".join(out_keys))
    plt.legend(vocab_out.keys, loc="best")

    assert (
        np.mean(
            similarity(
                sim.data[out_p][t_item1], vocab_out.parse(out_keys[0]).v, normalize=True
            )
        )
        > 0.9
    )
    assert (
        np.mean(
            similarity(
                sim.data[out_p][t_item2], vocab_out.parse(out_keys[1]).v, normalize=True
            )
        )
        > 0.9
    )



"""Caching capabilities for a faster build process."""

import errno
import hashlib
import logging
import os
import pathlib
import pickle
import struct
import warnings
from subprocess import CalledProcessError
from uuid import uuid1

import numpy as np

from nengo.dists import (
    PDF,
    Choice,
    CosineSimilarity,
    Exponential,
    Gaussian,
    Samples,
    ScatteredHypersphere,
    SqrtBeta,
    SubvectorLength,
    Uniform,
    UniformHypersphere,
)
from nengo.exceptions import (
    CacheIOError,
    CacheIOWarning,
    FingerprintError,
    TimeoutError,
)
from nengo.neurons import (
    LIF,
    AdaptiveLIF,
    AdaptiveLIFRate,
    Direct,
    Izhikevich,
    LIFRate,
    PoissonSpiking,
    RectifiedLinear,
    RegularSpiking,
    Sigmoid,
    SpikingRectifiedLinear,
    StochasticSpiking,
    Tanh,
)
from nengo.rc import rc
from nengo.solvers import (
    Lstsq,
    LstsqDrop,
    LstsqL1,
    LstsqL2,
    LstsqL2nz,
    LstsqMultNoise,
    LstsqNoise,
    Nnls,
    NnlsL2,
    NnlsL2nz,
)
from nengo.utils import nco
from nengo.utils.cache import byte_align, bytes2human, human2bytes
from nengo.utils.least_squares_solvers import (
    SVD,
    BlockConjgrad,
    Cholesky,
    Conjgrad,
    ConjgradScipy,
    LSMRScipy,
    RandomizedSVD,
)
from nengo.utils.lock import FileLock

logger = logging.getLogger(__name__)


def get_fragment_size(path):
    """Get fragment size in cross-compatible way."""
    try:
        return os.statvfs(path).f_frsize
    except AttributeError:  # pragma: no cover
        # no statvfs on Windows
        return 4096  # correct value in 99% of cases


def safe_stat(path):
    """Gets file stat, but fails gracefully in case of an OSError."""
    try:
        return path.stat()
    except OSError as err:
        logger.warning("OSError during safe_stat: %s", err)
    return None


def safe_remove(path):
    """Removes file, but fails gracefully in case of an OSError."""
    try:
        path.unlink()
    except OSError as err:  # pragma: no cover
        logger.warning("OSError during safe_remove: %s", err)


def safe_makedirs(path):
    """Try to make directories, but continue on error."""
    if not path.exists():
        try:
            path.mkdir(parents=True)
        except OSError as err:  # pragma: no cover
            logger.warning("OSError during safe_makedirs: %s", err)


def check_dtype(ndarray):
    """Check that array is a standard dtype."""
    return ndarray.dtype.isbuiltin == 1 and not ndarray.dtype.hasobject


def check_seq(tpl):
    """Check that all objects in list are fingerprintable."""
    return all(Fingerprint.supports(x) for x in tpl)


def check_mapping(mapping):
    """Check that all values in dict are fingerprintable."""
    return all(
        isinstance(key, str) and Fingerprint.supports(val)
        for key, val in mapping.items()
    )


def check_attrs(obj):
    """Check that all attributes of ``obj`` are fingerprintable."""
    attrs = [getattr(obj, x) for x in dir(obj) if not x.startswith("_")]
    return all(Fingerprint.supports(x) for x in attrs if not callable(x))


class Fingerprint:
    """Fingerprint of an object instance.

    A finger print is equal for two instances if and only if they are of the
    same type and have the same attributes.

    The fingerprint will be used as identification for caching.

    Parameters
    ----------
    obj : object
        Object to fingerprint.

    Attributes
    ----------
    fingerprint : hash
        A unique fingerprint for the object instance.

    Notes
    -----
    Not all objects can be fingerprinted. In particular, custom classes are
    tricky to fingerprint as their implementation can change without changing
    its fingerprint, as the type and attributes may be the same.

    In order to ensure that only safe object are fingerprinted, this class
    maintains class attribute ``WHITELIST`` that contains all types that can
    be safely fingerprinted.

    If you want your custom class to be fingerprinted, call the
    `.whitelist` class method and pass in your class.
    """

    __slots__ = ("fingerprint",)

    SOLVERS = (
        Lstsq,
        LstsqDrop,
        LstsqL1,
        LstsqL2,
        LstsqL2nz,
        LstsqNoise,
        LstsqMultNoise,
        Nnls,
        NnlsL2,
        NnlsL2nz,
    )
    LSTSQ_METHODS = (
        BlockConjgrad,
        Cholesky,
        Conjgrad,
        ConjgradScipy,
        LSMRScipy,
        RandomizedSVD,
        SVD,
    )
    NEURON_TYPES = (
        AdaptiveLIF,
        AdaptiveLIFRate,
        Direct,
        Izhikevich,
        LIF,
        LIFRate,
        PoissonSpiking,
        RectifiedLinear,
        RegularSpiking,
        Sigmoid,
        SpikingRectifiedLinear,
        StochasticSpiking,
        Tanh,
    )
    DISTRIBUTIONS = (
        Choice,
        CosineSimilarity,
        Exponential,
        Gaussian,
        PDF,
        Samples,
        ScatteredHypersphere,
        SqrtBeta,
        SubvectorLength,
        Uniform,
        UniformHypersphere,
    )

    WHITELIST = set(
        (
            type(None),
            bool,
            float,
            complex,
            bytes,
            list,
            tuple,
            np.ndarray,
            int,
            str,
            dict,
        )
        + SOLVERS
        + LSTSQ_METHODS
        + NEURON_TYPES
        + DISTRIBUTIONS
    )
    CHECKS = dict(
        [
            (np.ndarray, check_dtype),
            (tuple, check_seq),
            (list, check_seq),
            (dict, check_mapping),
        ]
        + [(_x, check_attrs) for _x in SOLVERS + LSTSQ_METHODS + NEURON_TYPES]
    )

    def __init__(self, obj):
        if not self.supports(obj):
            raise FingerprintError(
                f"Object of type '{type(obj).__name__}' cannot be fingerprinted."
            )

        self.fingerprint = hashlib.sha1()
        try:
            self.fingerprint.update(pickle.dumps(obj, pickle.HIGHEST_PROTOCOL))
        except Exception as err:  # pragma: no cover
            raise FingerprintError() from err

    def __str__(self):
        return self.fingerprint.hexdigest()

    @classmethod
    def supports(cls, obj):
        """Determines whether ``obj`` can be fingerprinted.

        Uses the `.whitelist`  method and runs the check function associated
        with the type of ``obj``.
        """
        typ = type(obj)
        in_whitelist = typ in cls.WHITELIST
        succeeded_check = typ not in cls.CHECKS or cls.CHECKS[typ](obj)
        return in_whitelist and succeeded_check

    @classmethod
    def whitelist(cls, typ, fn=None):
        """Whitelist the type given in ``typ``.

        Will run the check function ``fn`` on objects if provided.
        """
        cls.WHITELIST.add(typ)
        if fn is not None:
            cls.CHECKS[typ] = fn


class CacheIndex:
    """Cache index mapping keys to (filename, start, end) tuples.

    Once instantiated the cache index has to be used in a ``with`` block to
    allow access. The index will not be loaded before the ``with`` block is
    entered.

    This class only provides read access to the cache index. For write
    access use `.WriteableCacheIndex`.

    Examples
    --------

    .. testsetup::

       import tempfile

       tmpdir = tempfile.TemporaryDirectory()
       cache_dir = tmpdir.name

    .. testcode::

       from nengo.cache import CacheIndex, WriteableCacheIndex

       to_cache = ("gotta cache 'em all", 151)

       # create index file
       with WriteableCacheIndex(cache_dir) as index:
           index[hash(to_cache)] = ("file1", 0, 1)  # set an item

       # read from index
       with CacheIndex(cache_dir) as index:
           filename, start, end = index[hash(to_cache)]

    .. testcleanup::

       tmpdir.cleanup()

    Parameters
    ----------
    cache_dir : str
        Path where the cache is stored.

    Attributes
    ----------
    cache_dir : str
        Path where the cache is stored.
    index_path : str
        Path to the cache index file.
    version : tuple
        Version code of the loaded cache index. The first element gives the
        format of the cache and the second element gives the pickle protocol
        used to store the index. Note that a cache index will always be written
        in the newest version with the highest pickle protocol.
    VERSION (class attribute) : int
        Highest supported version, and version used to store the cache index.

    Notes
    -----
    Under the hood, the cache index is stored as a pickle file. The pickle file
    contains two objects which are read sequentially: the ``version`` tuple,
    and the ``index`` dictionary mapping keys to (filename, start, end) tuples.
    """

    _INDEX = "index"
    VERSION = 2

    def __init__(self, cache_dir):
        self.cache_dir = pathlib.Path(cache_dir)
        self.version = None
        self._index = None

    @property
    def index_path(self):
        return self.cache_dir / self._INDEX

    def __contains__(self, key):
        return key in self._index

    def __getitem__(self, key):
        return self._index[key]

    def __setitem__(self, key, value):
        raise TypeError("Index is readonly.")

    def __delitem__(self, key):
        raise TypeError("Index is readonly.")

    def __enter__(self):
        self._load_index_file()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _load_index_file(self):
        with self.index_path.open("rb") as f:
            self.version = pickle.load(f)
            if (
                self.version[0] > self.VERSION
                or self.version[1] > pickle.HIGHEST_PROTOCOL
            ):
                raise CacheIOError("Unsupported cache index file format.")
            self._index = pickle.load(f)
        assert isinstance(self.version, tuple)
        assert isinstance(self._index, dict)


class WriteableCacheIndex(CacheIndex):
    """Writable cache index mapping keys to files.

    This class allows write access to the cache index.

    The updated cache file will be written when the ``with`` block is exited.
    The initial read and the write on exit of the ``with`` block are locked
    against concurrent access with a file lock. The lock will be released
    within the ``with`` block.

    Examples
    --------

    .. testsetup::

       import tempfile

       tmpdir = tempfile.TemporaryDirectory()
       cache_dir = tmpdir.name

    .. testcode::

       from nengo.cache import WriteableCacheIndex

       to_cache = ("gotta cache 'em all", 151)
       key = hash(to_cache)
       with WriteableCacheIndex(cache_dir) as index:
           index[key] = ("file1", 0, 1)  # set an item
           del index[key]  # remove an item by key
           index.remove_file_entry("file1")  # remove an item by filename

    .. testcleanup::

       tmpdir.cleanup()

    Parameters
    ----------
    cache_dir : str
        Path where the cache is stored.
    """

    def __init__(self, cache_dir):
        super().__init__(cache_dir)
        self._lock = FileLock(str(self.index_path) + ".lock")
        self._updates = {}
        self._deletes = set()
        self._removed_files = set()

    def __getitem__(self, key):
        if key in self._updates:
            return self._updates[key]
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        if not isinstance(value, tuple) or len(value) != 3:
            raise ValueError("Cache entries must include filename, start, and end.")
        self._updates[key] = value

    def __delitem__(self, key):
        self._deletes.add(key)

    def __enter__(self):
        with self._lock:
            try:
                self._load_index_file()
            except (FileNotFoundError, EOFError):
                logger.exception("Decoder cache index corrupted. Reinitializing cache.")
                # If we can't load the index file, the cache is corrupted,
                # so we invalidate it (delete all files in the cache)
                self._reinit()
                self._write_index()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.sync()

    def _reinit(self):
        lock_filepath = pathlib.Path(self._lock.filename)

        def clear_directory(dir_path):
            for path in dir_path.iterdir():
                if path == lock_filepath:
                    continue
                if path.is_dir():
                    clear_directory(path)
                    path.rmdir()
                else:
                    path.unlink()

        clear_directory(self.cache_dir)
        self._index = {}

    def remove_file_entry(self, filename):
        """Remove entries mapping to ``filename``."""
        filename = pathlib.Path(filename)
        if self.cache_dir in filename.parents:
            filename = filename.relative_to(self.cache_dir)
        self._removed_files.add(str(filename))

    def _write_index(self):
        assert self._lock.acquired
        part_path = self.index_path.with_suffix(self.index_path.suffix + ".part")
        with part_path.open("wb") as f:
            # Use protocol 2 for version information to ensure that
            # all Python versions supported by Nengo will be able to
            # read it in the future.
            pickle.dump((self.VERSION, pickle.HIGHEST_PROTOCOL), f, 2)
            # Use highest available protocol for index data for maximum
            # performance.
            pickle.dump(self._index, f, pickle.HIGHEST_PROTOCOL)

        try:
            part_path.replace(self.index_path)
        except (CalledProcessError, PermissionError):
            # It may fail when
            # another program like a virus scanner is accessing the file to be
            # moved. There is not a lot we could do about this. See
            # <https://github.com/nengo/nengo/issues/1200> for more info.
            warnings.warn(
                "The cache index could not be updated because another program "
                "blocked access to it. This is commonly caused by anti-virus "
                "software. It is safe to ignore this warning. But if you see "
                "it a lot, you might want to consider doing one of the "
                "following for the best Nengo performance:\n"
                "1. Configure your anti-virus to ignore the Nengo cache "
                f"folder ('{self.cache_dir}').\n"
                "2. Disable the cache.\n",
                category=CacheIOWarning,
            )

    def sync(self):
        """Write changes to the cache index back to disk.

        The call to this function will be locked by a file lock.
        """
        try:
            with self._lock:
                try:
                    self._load_index_file()
                except IOError as err:  # pragma: no cover
                    if err.errno == errno.ENOENT:
                        self._index = {}
                    else:
                        raise

                self._index.update(self._updates)
                for key in self._deletes:
                    del self._index[key]
                self._index = {
                    k: v
                    for k, v in self._index.items()
                    if v[0] not in self._removed_files
                }

                self._write_index()
        except TimeoutError:  # pragma: no cover
            warnings.warn(
                "Decoder cache index could not acquire lock. "
                "Cache index was not synced."
            )

        self._updates.clear()
        self._deletes.clear()
        self._removed_files.clear()


class DecoderCache:
    """Cache for decoders.

    Hashes the arguments to the decoder solver and stores the result in a file
    which will be reused in later calls with the same arguments.

    Be aware that decoders should not use any global state, but only values
    passed and attributes of the object instance. Otherwise the wrong solver
    results might get loaded from the cache.

    Parameters
    ----------
    readonly : bool
        Indicates that already existing items in the cache will be used, but no
        new items will be written to the disk in case of a cache miss.
    cache_dir : str or None
        Path to the directory in which the cache will be stored. It will be
        created if it does not exists. Will use the value returned by
        `.get_default_dir`, if ``None``.
    """

    _CACHE_EXT = ".nco"

    def __init__(self, readonly=False, cache_dir=None):
        self.readonly = readonly
        if cache_dir is None:
            cache_dir = self.get_default_dir()
        self.cache_dir = pathlib.Path(cache_dir)
        if readonly:
            self._index = CacheIndex(self.cache_dir)
        else:
            safe_makedirs(self.cache_dir)
            self._index = WriteableCacheIndex(self.cache_dir)
        self._fragment_size = get_fragment_size(self.cache_dir)
        self._fd = None
        self._in_context = False

    def __enter__(self):
        try:
            try:
                self._index.__enter__()
            except TimeoutError:  # pragma: no cover
                self.readonly = True
                self._index = CacheIndex(self.cache_dir)
                self._index.__enter__()
                warnings.warn(
                    "Decoder cache could not acquire lock and was "
                    "set to readonly mode."
                )
        except Exception as e:  # pylint: disable=broad-except
            self.readonly = True
            self._index = None
            logger.debug("Could not acquire lock because: %s", e)
            warnings.warn("Decoder cache could not acquire lock and was deactivated.")
        self._in_context = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._in_context = False
        self._close_fd()
        if self._index is not None:
            return self._index.__exit__(exc_type, exc_value, traceback)
        return None

    @staticmethod
    def get_default_dir():
        """Returns the default location of the cache.

        Returns
        -------
        str
        """
        return rc["decoder_cache"]["path"]

    def _close_fd(self):
        if self._fd is not None:
            self._fd.close()
            self._fd = None

    def _get_fd(self):
        if self._fd is None:
            self._fd = open(  # pylint: disable=consider-using-with
                self._key2path(str(uuid1())), "wb"
            )
        return self._fd

    def get_files(self):
        """Returns all of the files in the cache.

        Returns
        -------
        list of (str, int) tuples
        """
        files = []
        for path in self.cache_dir.iterdir():
            if path.is_dir():
                files.extend(path.iterdir())
        return files

    def get_size(self):
        """Returns the size of the cache with units as a string.

        Returns
        -------
        str
        """
        return bytes2human(self.get_size_in_bytes())

    def get_size_in_bytes(self):
        """Returns the size of the cache in bytes as an int.

        Returns
        -------
        int
        """
        stats = (safe_stat(f) for f in self.get_files())
        return sum(
            byte_align(st.st_size, self._fragment_size)
            for st in stats
            if st is not None
        )

    def invalidate(self):
        """Invalidates the cache (i.e. removes all cache files)."""
        if self.readonly:
            raise CacheIOError("Cannot invalidate a readonly cache.")
        self._close_fd()
        with self._index:
            for path in self.get_files():
                self.remove_file(path)

    def shrink(self, limit=None):  # noqa: C901
        """Reduces the size of the cache to meet a limit.

        Parameters
        ----------
        limit : int, optional
            Maximum size of the cache in bytes.
        """
        if self.readonly:
            logger.info("Tried to shrink a readonly cache.")
            return

        if limit is None:
            limit = rc["decoder_cache"]["size"]
        if isinstance(limit, str):
            limit = human2bytes(limit)

        self._close_fd()

        fileinfo = []
        excess = -1 * limit
        for path in self.get_files():
            stat = safe_stat(path)
            if stat is not None:
                aligned_size = byte_align(stat.st_size, self._fragment_size)
                excess += aligned_size
                fileinfo.append((stat.st_atime, aligned_size, path))

        # Remove the least recently accessed first
        fileinfo.sort()

        try:
            with self._index:
                for _, size, path in fileinfo:
                    if excess <= 0:
                        break

                    excess -= size
                    self.remove_file(path)
        except TimeoutError:  # pragma: no cover
            logger.debug("Not shrinking cache. Lock could not be acquired.")

    def remove_file(self, path):
        """Removes the file at ``path`` from the cache."""
        self._index.remove_file_entry(path)
        safe_remove(path)

    def wrap_solver(self, solver_fn):  # noqa: C901
        """Takes a decoder solver and wraps it to use caching.

        Parameters
        ----------
        solver : func
            Decoder solver to wrap for caching.

        Returns
        -------
        func
            Wrapped decoder solver.
        """

        def cached_solver(
            conn, gain, bias, x, targets, rng=np.random, **uncached_kwargs
        ):
            if not self._in_context or self._index is None:
                if not self._in_context:
                    warnings.warn(
                        "Cannot use cached solver outside of `with cache` block."
                    )
                return solver_fn(
                    conn, gain, bias, x, targets, rng=rng, **uncached_kwargs
                )

            try:
                key = self._get_cache_key(
                    conn.solver, conn.pre_obj.neuron_type, gain, bias, x, targets, rng
                )
            except FingerprintError as e:  # pragma: no cover
                logger.debug("Failed to generate cache key: %s", e)
                return solver_fn(
                    conn, gain, bias, x, targets, rng=rng, **uncached_kwargs
                )

            try:
                path, start, end = self._index[key]
                if self._fd is not None:
                    self._fd.flush()
                with open(path, "rb") as f:
                    f.seek(start)
                    info, decoders = nco.read(f)
            except (KeyError, FileNotFoundError, struct.error) as err:
                if isinstance(err, KeyError):
                    logger.debug("Cache miss [%s].", key)
                else:
                    logger.exception("Corrupted cache entry [%s].", key)
                decoders, info = solver_fn(
                    conn, gain, bias, x, targets, rng=rng, **uncached_kwargs
                )
                if not self.readonly:
                    fd = self._get_fd()
                    start = fd.tell()
                    nco.write(fd, info, decoders)
                    end = fd.tell()
                    self._index[key] = (fd.name, start, end)
            else:
                logger.debug("Cache hit [%s]: Loaded stored decoders.", key)
            return decoders, info

        return cached_solver

    def _get_cache_key(self, solver, neuron_type, gain, bias, x, targets, rng):
        h = hashlib.sha1()

        h.update(str(Fingerprint(solver)).encode("utf-8"))
        h.update(str(Fingerprint(neuron_type)).encode("utf-8"))

        h.update(np.ascontiguousarray(gain).data)
        h.update(np.ascontiguousarray(bias).data)
        h.update(np.ascontiguousarray(x).data)
        h.update(np.ascontiguousarray(targets).data)

        # rng format doc:
        # https://numpy.org/doc/stable/reference/random/generated/numpy.random.RandomState.get_state.html#numpy.random.RandomState.get_state
        state = rng.get_state()
        h.update(state[0].encode())  # string 'MT19937'
        h.update(state[1].data)  # 1-D array of 624 unsigned integer keys
        h.update(struct.pack("q", state[2]))  # integer pos
        h.update(struct.pack("q", state[3]))  # integer has_gauss
        h.update(struct.pack("d", state[4]))  # float cached_gaussian

        return h.hexdigest()

    def _key2path(self, key):
        prefix = key[:2]
        suffix = key[2:]
        directory = self.cache_dir / prefix
        safe_makedirs(directory)
        return directory / (suffix + self._CACHE_EXT)


class NoDecoderCache:
    """Provides the same interface as `.DecoderCache` without caching."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def wrap_solver(self, solver_fn):
        return solver_fn

    def get_size_in_bytes(self):
        return 0

    def get_size(self):
        return "0 B"

    def shrink(self, limit=0):
        pass

    def invalidate(self):
        pass


def get_default_decoder_cache():
    """Get default decoder implementation based on config settings."""
    if rc["decoder_cache"].getboolean("enabled"):
        decoder_cache = DecoderCache(rc["decoder_cache"].getboolean("readonly"))
    else:
        decoder_cache = NoDecoderCache()
    return decoder_cache



"""Helper functions for constructing nengo models that implement actions."""
import numpy as np

import nengo


def convolution(module, target_name, effect, n_neurons_cconv, synapse):
    """Implement an action_objects.Convolution.

    Parameters
    ----------
    module : spa.Module
        The module that will own this convolution
    target_name : string
        The name of the object to send the convolution result to
    effect : action_objects.Convolution
        The details of the convolution to implement
    n_neurons_cconv : int
        Number of neurons in each product population
    synapse : float (or nengo.Synapse)
        The synapse to use for connections into and out of the convolution

    Returns the created nengo.networks.CircularConvolution.
    """

    source1 = effect.source1
    source2 = effect.source2

    target_module = module.spa.get_module(target_name)
    target, target_vocab = module.spa.get_module_input(target_name)
    s1_output, s1_vocab = module.spa.get_module_output(source1.name)
    s2_output, s2_vocab = module.spa.get_module_output(source2.name)

    with target_module:
        cconv = nengo.networks.CircularConvolution(
            n_neurons_cconv,
            s1_vocab.dimensions,
            invert_a=False,
            invert_b=False,
            label="cconv_%s" % str(effect),
        )

    with module.spa:
        # compute the requested transform
        t = s1_vocab.parse(str(effect.transform)).get_convolution_matrix()
        # handle conversion between different Vocabularies
        if target_vocab is not s1_vocab:
            t = np.dot(s1_vocab.transform_to(target_vocab), t)

        nengo.Connection(cconv.output, target, transform=t, synapse=synapse)

        t1 = s1_vocab.parse(source1.transform.symbol).get_convolution_matrix()
        if source1.inverted:
            D = s1_vocab.dimensions
            t1 = np.dot(t1, np.eye(D)[-np.arange(D)])

        nengo.Connection(s1_output, cconv.input_a, transform=t1, synapse=synapse)

        t2 = s2_vocab.parse(source2.transform.symbol).get_convolution_matrix()
        if source2.inverted:
            D = s2_vocab.dimensions
            t2 = np.dot(t2, np.eye(D)[-np.arange(D)])
        if s1_vocab is not s2_vocab:
            t2 = np.dot(s2_vocab.transform_to(s1_vocab), t2)
        nengo.Connection(s2_output, cconv.input_b, transform=t2, synapse=synapse)
    return cconv



def function_name(func):
    """Returns the name of a function.

    Unlike accessing ``func.__name__``, this function is robust to the
    different types of objects that can be considered a function in Nengo.

    Parameters
    ----------
    func : callable or array_like
        Object used as function argument.

    Returns
    -------
    str
        Name of function object.
    """
    return getattr(func, "__name__", func.__class__.__name__)



import numpy as np
import pytest

import nengo
from nengo.exceptions import BuildError, ValidationError


@pytest.mark.parametrize("use_dist", (False, True))
@pytest.mark.parametrize("use_scipy", (False, True))
def test_sparse(use_dist, use_scipy, Simulator, rng, seed, plt, monkeypatch, allclose):
    if use_scipy:
        scipy_sparse = pytest.importorskip("scipy.sparse")
    else:
        monkeypatch.setattr(nengo.transforms, "scipy_sparse", None)
        monkeypatch.setattr(nengo.utils.numpy, "scipy_sparse", None)
        monkeypatch.setattr(nengo.utils.numpy, "is_spmatrix", lambda obj: False)

    input_d = 4
    output_d = 2
    shape = (output_d, input_d)

    inds = np.asarray([[0, 0], [1, 1], [0, 2], [1, 3]])
    weights = rng.uniform(0.25, 0.75, size=4)
    if use_dist:
        init = nengo.dists.Uniform(0.25, 0.75)
        indices = inds
    elif use_scipy:
        init = scipy_sparse.csr_matrix((weights, inds.T), shape=shape)
        indices = None
    else:
        init = weights
        indices = inds

    transform = nengo.transforms.Sparse(shape, indices=indices, init=init)

    sim_time = 1.0
    with nengo.Network(seed=seed) as net:
        x = nengo.processes.WhiteSignal(period=sim_time, high=10, seed=seed + 1)
        u = nengo.Node(x, size_out=4)
        a = nengo.Ensemble(100, 2)
        conn = nengo.Connection(u, a, synapse=None, transform=transform)
        ap = nengo.Probe(a, synapse=0.03)

    def run_sim():
        with Simulator(net) as sim:
            sim.run(sim_time)
        return sim

    if use_scipy:
        sim = run_sim()
    else:
        with pytest.warns(UserWarning, match="require Scipy"):
            sim = run_sim()

    actual_weights = sim.data[conn].weights

    full_transform = np.zeros(shape)
    full_transform[inds[:, 0], inds[:, 1]] = weights
    if use_dist:
        actual_weights = actual_weights.toarray()
        assert np.array_equal(actual_weights != 0, full_transform != 0)
        full_transform[:] = actual_weights

    conn.transform = full_transform
    with Simulator(net) as ref_sim:
        ref_sim.run(sim_time)

    plt.plot(ref_sim.trange(), ref_sim.data[ap], ":")
    plt.plot(sim.trange(), sim.data[ap])

    assert allclose(sim.data[ap], ref_sim.data[ap])


@pytest.mark.parametrize("encoders", (True, False))
@pytest.mark.parametrize("decoders", (True, False))
def test_sparse_nef(encoders, decoders, Simulator):
    """Sparse transforms currently don't work with NEF connections."""

    with nengo.Network() as net:
        transform = nengo.transforms.Sparse((2, 2), indices=[[0, 1], [1, 0]])
        a = nengo.Ensemble(2, 2)
        b = nengo.Ensemble(2, 2)
        nengo.Connection(
            a if decoders else a.neurons,
            b if encoders else b.neurons,
            transform=transform,
        )

    if decoders:
        # error if decoders
        with pytest.raises(BuildError, match="decoded connection"):
            with Simulator(net):
                pass
    else:
        # no error
        with Simulator(net):
            pass


def test_sparse_validation_errors():
    with pytest.raises(ValidationError, match="Either `init` must be a `scipy.sparse"):
        nengo.Sparse((1, 1))

    nengo.Sparse((3, 3), indices=[(0, 0), (1, 2)], init=[1, 2])
    with pytest.raises(ValidationError, match="Must be a vector.*length as `ind"):
        nengo.Sparse((3, 3), indices=[(0, 0), (1, 2)], init=[1, 2, 3])


def test_sparseinitparam_errors():
    class TestClass:
        sparse = nengo.transforms.SparseInitParam("sparse")

    test = TestClass()
    with pytest.raises(ValidationError, match="Must be `.*SparseMatrix` or .*spmatrix"):
        test.sparse = "a"


def test_notransform():
    transform = nengo.transforms.NoTransform(3)
    assert transform.size_in == transform.size_out == 3

    with pytest.raises(TypeError, match="Cannot sample a NoTransform"):
        transform.sample()
